// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

//     **** This file must be preprocessed before compilation ****

// RULE: coordinates on non-static methods are ALWAYS relative to the
// upper-left hand corner of <tt>this</tt>

// FEATURE: reflow before allowing js to read from width/height 
// FEATURE: fastpath for rows=1/cols=1
// FEATURE: reflow starting with a certain child
// FEATURE: separate mark_for_reflow and mark_for_resize

import java.io.*;
import java.net.*;
import java.util.*;
import org.xwt.js.*;
import org.xwt.util.*;
import org.xwt.translators.*;

/**
 *  <p>
 *  Encapsulates the data for a single XWT box as well as all layout
 *  rendering logic.
 *  </p>
 *
 *  <p>
 *  This is the real meat of XWT. Part of its monolithic design is for
 *  performance reasons: deep inheritance heirarchies are slow, and
 *  neither javago nor GCJ can inline across class boundaries.
 *  </p>
 *
 *  <p>The rendering process consists of three phases; each requires
 *     one DFS pass over the tree</p>
 *
 *  <ol><li> <b>repacking</b>: children of a box are packed into columns
 *           and rows according to their colspan/rowspan attributes and
 *           ordering.  Minimum and maximum sizes of columns are computed.
 *
 *      <li> <b>resizing</b>: width/height and x/y positions of children
 *           are assigned.  If a PosChange or SizeChange is triggered,
 *           <tt>Surface.abort</tt> will be set and the resizing process will
 *           Surface.abort.
 *
 *      <li> <b>repainting</b>: children draw their content onto the
 *           buffer.
 *  </ol>
 *
 *  The first two passes together are called the <i>reflow</i> phase.
 *
 *  Reflowing is done in a seperate pass since PosChanges and
 *  SizeChanges trigger an Surface.abort; if rendering were done in the same
 *  pass, rendering work done prior to the Surface.abort would be wasted.
 *
 *  The two passes are repack and resize.  Together they are known as
 *  reflow.  Repacking assigns boxes to the appropriate grid
 *  coordinates within the parents and computes
 *  contentwidth/contentheight.  Resize computes actual size and
 *  position.
 *
 *  A note on coordinates: the Box class represents regions
 *  internally as x,y,w,h tuples, even though the PixelBuffer class
 *  uses x1,y1,x2,y2 tuples.
 */
public final class Box extends JS.Scope {

    public Box() { super(null); }


    // Misc instance data ////////////////////////////////////////////////////////////////

    static int sizePosChangesSinceLastRender = 0;


    // Misc instance data ////////////////////////////////////////////////////////////////

    boolean needs_reflow = true;         
    //#define MARK_FOR_REFLOW_this for(Box b2 = this; b2 != null && !b2.needs_reflow; b2 = b2.parent) b2.needs_reflow = true;
    //#define MARK_FOR_REFLOW_b for(Box b2 = b; b2 != null && !b2.needs_reflow; b2 = b2.parent) b2.needs_reflow = true;
    //#define MARK_FOR_REFLOW_b_parent for(Box b2 = b.parent; b2 != null && !b2.needs_reflow; b2 = b2.parent) b2.needs_reflow = true;

    Box redirect = this;
    Surface surface = null;               // null on all non-root boxen

    // Flags ///////////////////////////////////////////////////////////////////////////////
    short flags = 0;
    static int MOUSEINSIDE_FLAG  = 0x00000001;
    static int INVISIBLE_FLAG    = 0x00000002;
    static int ABSOLUTE_FLAG     = 0x00000004;
    static int HSHRINK_FLAG      = 0x00000008;
    static int VSHRINK_FLAG      = 0x00000010;
    static int TILE_FLAG         = 0x00000020;
    static int FONT_CHANGED_FLAG = 0x00000040;  // set when font changes, cleared during repack
    static int ISROOT_FLAG       = 0x00000080;


    // Geometry ////////////////////////////////////////////////////////////////////////////

    // xwt can be compiled with 16-bit lengths to save memory on small devices
    //#define LENGTH int
    //#define MAX_LENGTH Integer.MAX_VALUE
    //#define MIN_LENGTH Integer.MIN_VALUE

    // always correct (set directly by user)
    public LENGTH minwidth = 0;        
    public LENGTH minheight = 0;        
    public LENGTH maxwidth = MAX_LENGTH;
    public LENGTH maxheight = MAX_LENGTH;
    private LENGTH hpad = 0;
    private LENGTH vpad = 0;
    private String text = null;
    private Res font = null;
    private int fontsize = 10;
    private LENGTH textwidth = 0;
    private LENGTH textheight = 0;

    // FEATURE: use shorts
    private int rows = 1;
    private int cols = 0;
    private int rowspan = 1;
    private int colspan = 1;

    // computed during reflow
    public LENGTH x = 0;
    public LENGTH y = 0;
    public LENGTH width = 0;
    public LENGTH height = 0;
    private int row = 0;  // FEATURE use a short
    private int col = 0;  // FEATURE use a short
    private LENGTH contentwidth = 0;             // == max(minwidth, textwidth+pad, sum(child.contentwidth) + pad)
    private LENGTH contentheight = 0;


    // Rendering Properties ///////////////////////////////////////////////////////////

    //private SVG.VP path = null;
    //private SVG.Paint fill = null;
    //private SVG.Paint stroke = null;

    public Picture image;                        // will disappear
    private int fillcolor = 0x00000000;          // will become SVG.Paint
    private int strokecolor = 0xFF000000;        // will become SVG.Paint

    private String cursor = null;                // the cursor for this box


    // Instance Methods /////////////////////////////////////////////////////////////////////

    /** Adds the intersection of (x,y,w,h) and the node's current actual geometry to the Surface's dirty list */
    public final void dirty() { dirty(0, 0, width, height); }
    public final void dirty(int x, int y, int w, int h) {
        for(Box cur = this; cur != null; cur = cur.parent) {
            w = min(x + w, cur.width) - max(x, 0);
            h = min(y + h, cur.height) - max(y, 0);
            x = max(x, 0);
            y = max(y, 0);
            if (w <= 0 || h <= 0) return;
            if (cur.parent == null && cur.surface != null) cur.surface.dirty(x, y, w, h);
            x += cur.x;
            y += cur.y;
        }
    }

    /**
     *  Given an old and new mouse position, this will update <tt>mouseinside</tt> and check
     *  to see if this node requires any Enter, Leave, or Move notifications.
     *
     *  @param forceleave set to true by the box's parent if the mouse is inside an older
     *                    sibling, which is covering up this box.
     */
    void Move(int oldmousex, int oldmousey, int mousex, int mousey) { Move(oldmousex, oldmousey, mousex, mousey, false); }
    void Move(int oldmousex, int oldmousey, int mousex, int mousey, boolean forceleave) {

        boolean wasinside = (flags & MOUSEINSIDE_FLAG) != 0;
        boolean isinside = !((flags & INVISIBLE_FLAG) != 0) && inside(mousex, mousey) && !forceleave;
        if (isinside) flags |= MOUSEINSIDE_FLAG; else flags &= ~MOUSEINSIDE_FLAG;

        if (!wasinside && !isinside) return;

        if (!wasinside && isinside && get("Enter", Trap.class) != null) put("Enter", Boolean.TRUE);
        else if (wasinside && !isinside && get("Leave", Trap.class) != null) put("Leave", Boolean.TRUE);
        else if (wasinside && isinside && (mousex != oldmousex || mousey != oldmousey) && get("Move", Trap.class) != null)
            put("Move", Boolean.TRUE);

        if (isinside && cursor != null) getRoot().cursor = cursor;

        // if the mouse has moved into our padding region, it is considered 'outside' all our children
        if (!(mousex >= hpad && mousey >= vpad && mousex < width - hpad && mousey < height + vpad)) forceleave = true;

        for(Box b = getChild(numChildren() - 1); b != null; b = b.prevSibling()) {
            b.Move(oldmousex - b.x, oldmousey - b.y, mousex - b.x, mousey - b.y, forceleave);
            if (b.inside(mousex - b.x, mousey - b.y)) forceleave = true;
        }
    }


    // Reflow ////////////////////////////////////////////////////////////////////////////////////////

    void reflow() { reflow(width, height); }
    void reflow(int new_width, int new_height) {
        repack();
        if (Surface.abort) return;
        resize(x, y, new_width, new_height);
    }

    /** Checks if the Box's size has changed, dirties it if necessary, and makes sure childrens' sizes are up to date */
    void repack() {
        if (!needs_reflow) return;
        if (numChildren() == 0) {
	    contentwidth = max(textwidth + 2 * hpad, minwidth);
	    contentheight = max(textheight + 2 * vpad, minheight);
	    return;
	}

        // --- Phase 0 ----------------------------------------------------------------------
        // recurse
        for(Box child = getChild(0); child != null; child = child.nextSibling()) {
            if (((flags & FONT_CHANGED_FLAG) != 0) && child.font == null) {
                child.flags |= FONT_CHANGED_FLAG;
                child.needs_reflow = true;
            }
            child.repack();
            if (Surface.abort) { MARK_FOR_REFLOW_this; return; }
        }
        flags &= ~FONT_CHANGED_FLAG;

        // --- Phase 1 ----------------------------------------------------------------------
        // assign children to their row/column positions (assuming constrained columns)
        if ((rows == 0 && cols == 0) || (rows != 0 && cols != 0)) throw new Error("rows == " + rows + "   cols == " + cols);
        //#repeat x/y y/x width/height col/row row/col cols/rows rows/cols colspan/rowspan rowspan/colspan colWidth/rowHeight numRowsInCol/numColsInRow INNER/INNER2 maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight OUTER/OUTER2 INNER/INNER2
        if (rows == 0) {
            int[] numRowsInCol = new int[cols];           // the number of cells occupied in each column
            Box child = getChild(0);
            for(; child != null && (((child.flags & ABSOLUTE_FLAG) != 0) || ((child.flags & INVISIBLE_FLAG) != 0)); child = child.nextSibling());
            OUTER: for(int row=0; child != null; row++) {
                for(int col=0; child != null && col < cols;) {
                    INNER: while(true) {  // scan across the row, looking for an unoccupied gap at least as wide as the child
                        while(col < cols && numRowsInCol[col] > row) col++;
                        for(int i=col; i < cols && i < col + min(cols, child.colspan); i++)
                            if (numRowsInCol[col] > row) { col = i + 1; continue INNER; }
                        break;
                    }
                    if (col + min(cols, child.colspan) > cols) break;
                    for(int i=col; i < col + min(cols, child.colspan); i++) numRowsInCol[i] += child.rowspan;
                    child.col = col;
                    child.row = row;
                    col += min(cols, child.colspan);
                    child = child.nextSibling();
                    for(; child != null && (((child.flags & ABSOLUTE_FLAG) != 0) || ((child.flags & INVISIBLE_FLAG) != 0)); child = child.nextSibling());
                }
            }
        }
        //#end

        // --- Phase 2 ----------------------------------------------------------------------
        // compute the min/max sizes of the columns and rows and set our contentwidth
        //#repeat x/y y/x width/height col/row cols/rows colspan/rowspan colWidth/rowHeight maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight numCols/numRows hpad/vpad
        contentwidth = 2 * hpad;
        int numCols = cols;
        if (numCols == 0)
            for(Box child = getChild(0); child != null; child = child.nextSibling())
                numCols = max(numCols, child.col + child.colspan);
        LENGTH[] colWidth = new LENGTH[numCols];
        for(Box child = getChild(0); child != null; child = child.nextSibling())
            if (!(((child.flags & ABSOLUTE_FLAG) != 0) || ((child.flags & INVISIBLE_FLAG) != 0)))
                colWidth[child.col] = max(colWidth[child.col], child.contentwidth / child.colspan);
        for(int col=0; col<numCols; col++) contentwidth += colWidth[col];
        contentwidth = max(textwidth + 2 * hpad, contentwidth);
        contentwidth = bound(minwidth, contentwidth, maxwidth);
        //#end
    }


    private void resize(LENGTH x, LENGTH y, LENGTH width, LENGTH height) {
        
        // --- Phase 1 ----------------------------------------------------------------------
        // run PosChange/SizeChange, dirty as needed
        if (x != this.x || y != this.y || width != this.width || height != this.height) {
            (parent == null ? this : parent).dirty(this.x, this.y, this.width, this.height);
            boolean sizechange = false, poschange = false;
            if ((this.width != width || this.height != height) && get("SizeChange", Trap.class) != null) sizechange = true;
            if ((this.x != x || this.y != y) && get("PosChange", Trap.class) != null) poschange = true;
            this.width = width; this.height = height; this.x = x; this.y = y;
            dirty();
            if (sizechange || poschange)
                if (sizePosChangesSinceLastRender == 500) {
                    if (Log.on) Log.logJS(this, "Warning, more than 500 SizeChange/PosChange traps triggered since last complete render");
                } else {
                    sizePosChangesSinceLastRender++;
                    try { if (sizechange) put("SizeChange", Boolean.TRUE); } catch (Exception e) { Log.log(this, e); }
                    try { if (poschange) put("PosChange", Boolean.TRUE); } catch (Exception e) { Log.log(this, e); }
                    Surface.abort = true;
                    return;
                }
            needs_reflow = true;
        }

        // --- short circuit ----------------------------------------------------------------
        if (!needs_reflow) return;
        needs_reflow = false;
        if (numChildren() == 0) return;

        // --- Phase 2 ----------------------------------------------------------------------
        // compute the min/max sizes of the columns and rows and set initial width/height to minimums

        //#repeat x/y y/x width/height col/row cols/rows colspan/rowspan colWidth/rowHeight maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight marginWidth/marginHeight numCols/numRows
        int numCols = cols;
        if (numCols == 0)
            for(Box child = getChild(0); child != null; child = child.nextSibling())
                numCols = max(numCols, child.col + child.colspan);
        LENGTH[] colWidth = new LENGTH[numCols];
        LENGTH[] colMaxWidth = new LENGTH[numCols];
        int marginWidth = width;
        for(int i=0; i<colMaxWidth.length; i++) colMaxWidth[i] = -1;
        //#end

        for(Box child = getChild(0); child != null; child = child.nextSibling()) {
            if (((child.flags & ABSOLUTE_FLAG) != 0) || ((child.flags & INVISIBLE_FLAG) != 0)) continue;
            //#repeat x/y y/x width/height col/row cols/rows colspan/rowspan colWidth/rowHeight maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight HSHRINK_FLAG/VSHRINK_FLAG numCols/numRows
            colWidth[child.col] = max(colWidth[child.col], child.contentwidth / child.colspan);
            for(int i=child.col; i<child.col+child.colspan && i<numCols; i++)
                colMaxWidth[i] = max(colMaxWidth[i], (((child.flags & HSHRINK_FLAG) != 0) ? child.contentwidth : child.maxwidth) / child.colspan);
            //#end
        }

        //#repeat x/y y/x width/height col/row cols/rows colspan/rowspan colWidth/rowHeight maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight marginWidth/marginHeight
        for(int i=0; i<colMaxWidth.length; i++) if (colMaxWidth[i] == -1) colMaxWidth[i] = MAX_LENGTH;

        for(int i=0; i<colMaxWidth.length; i++) {
            if (colMaxWidth[i] == MAX_LENGTH) { marginWidth = 0; break; }
            marginWidth -= colMaxWidth[i];
            if (marginWidth < 0) { marginWidth = 0; break; }
        }
        //#end
      

        // --- Phase 3 ----------------------------------------------------------------------
        // hand out the slack
        int slack;
        //#repeat x/y y/x width/height col/row cols/rows colspan/rowspan colWidth/rowHeight maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight numCols/numRows hpad/vpad
        slack = width - 2 * hpad;
        for(int i=0; i<numCols; i++) slack -= colWidth[i];
        if (numChildren() > 0)
            while(slack > 0) {  
                // FEATURE: inefficient
                int startslack = slack;
                int increment = max(1, slack / numCols);
                for(int col=0; col < numCols && slack > 0; col++) {
                    slack += colWidth[col];
                    colWidth[col] = min(colMaxWidth[col], colWidth[col] + increment);
                    slack -= colWidth[col];
                }
                if (slack == startslack) break;
            }   
        //#end


        // --- Phase 4 ----------------------------------------------------------------------
        // assign children's new sizes and positions and recurse
        for(Box child = getChild(0); child != null; child = child.nextSibling()) {
            if ((child.flags & INVISIBLE_FLAG) != 0) continue;
            int child_x = 0, child_y = 0, child_width = 0, child_height = 0;
            if ((child.flags & ABSOLUTE_FLAG) != 0) {
                child_x = child.x;
                child_y = child.y;
                child_width = ((child.flags & HSHRINK_FLAG) != 0) ? child.contentwidth : min(child.maxwidth, width - child.x - hpad);
                child_height = ((child.flags & VSHRINK_FLAG) != 0) ? child.contentheight : min(child.maxheight, height - child.y - vpad);
            } else {
                int diff;
                //#repeat x/y y/x width/height col/row cols/rows colspan/rowspan colWidth/rowHeight maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight HSHRINK_FLAG/VSHRINK_FLAG marginWidth/marginHeight hpad/vpad child_x/child_y child_width/child_height
                child_width = 0; for(int i=child.col; i<child.col+child.colspan && i<colWidth.length; i++) child_width += colWidth[i];
                diff = bound(child.contentwidth, child_width, ((child.flags & HSHRINK_FLAG) != 0) ? child.contentwidth : child.maxwidth) - child_width;
                child_x = max(hpad, marginWidth / 2); for(int i=0; i<child.col; i++) child_x += colWidth[i];
                if (diff < 0) child_x += -1 * (diff / 2);
                child_width += diff;
                //#end
            }
            child.resize(child_x, child_y, child_width, child_height);
        }
    }




    // Rendering Pipeline /////////////////////////////////////////////////////////////////////

    /** Renders self and children within the specified region. All rendering operations are clipped to xIn,yIn,wIn,hIn */
    void render(int parentx, int parenty, int clipx, int clipy, int clipw, int cliph, PixelBuffer buf) {
        if (Surface.abort || (flags & INVISIBLE_FLAG) != 0) return;
        int globalx = parentx + (parent == null ? 0 : x);
        int globaly = parenty + (parent == null ? 0 : y);

        // intersect the x,y,w,h rendering window with ourselves; quit if it's empty
        clipw = min(max(clipx, parent == null ? 0 : globalx) + clipw, (parent == null ? 0 : globalx) + width) - globalx;
        cliph = min(max(clipy, parent == null ? 0 : globaly) + cliph, (parent == null ? 0 : globaly) + height) - globaly;
        clipx = max(clipx, parent == null ? 0 : globalx);
        clipy = max(clipy, parent == null ? 0 : globaly);
        if (clipw <= 0 || cliph <= 0) return;

        if ((fillcolor & 0xFF000000) != 0x00000000)
            buf.fillRect(clipx, clipy, clipx + clipw, clipy + cliph, fillcolor);

        if (image != null)
            if ((flags & TILE_FLAG) != 0) renderTiledImage(globalx, globaly, clipx, clipy, clipw, cliph, buf);
            else renderStretchedImage(globalx, globaly, clipx, clipy, clipw, cliph, buf);

	if (text != null && !text.equals(""))
            renderText(globalx, globaly, clipx, clipy, clipw, cliph, buf);

        // now subtract the pad region from the clip region before proceeding
        clipw = min(max(clipx, globalx + hpad) + clipw, globalx + width - hpad) - clipx;
        cliph = min(max(clipy, globaly + vpad) + cliph, globaly + height - vpad) - clipy;
        clipx = max(clipx, globalx + hpad);
        clipy = max(clipy, globaly + vpad);

        for(Box b = getChild(0); b != null; b = b.nextSibling())
            b.render(globalx, globaly, clipx, clipy, clipw, cliph, buf);   
    }

    void renderStretchedImage(int globalx, int globaly, int clipx, int clipy, int clipw, int cliph, PixelBuffer buf) {
        // FIXME: wrong
        buf.drawPicture(image, clipx, clipy, clipx + clipw, clipy + cliph, 0, 0, image.getWidth(), image.getHeight());
    }

    void renderTiledImage(int globalx, int globaly, int x, int y, int w, int h, PixelBuffer buf) {
        int iw = image.getWidth();
        int ih = image.getHeight();
        for(int i=(x - x)/iw; i <= (x + w - x)/iw; i++) {
            for(int j=(y - y)/ih; j<= (y + h - y)/ih; j++) {
                
                int dx1 = max(i * iw + x, x);
                int dy1 = max(j * ih + y, y);
                int dx2 = min((i+1) * iw + x, x + w);
                int dy2 = min((j+1) * ih + y, y + h);
                
                int sx1 = dx1 - (i*iw) - x;
                int sy1 = dy1 - (j*ih) - y;
                int sx2 = dx2 - (i*iw) - x;
                int sy2 = dy2 - (j*ih) - y;

                if (dx2 - dx1 > 0 && dy2 - dy1 > 0 && sx2 - sx1 > 0 && sy2 - sy1 > 0)
                    buf.drawPicture(image, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2);
            }
        }
    }

    void renderText(int x, int y, int clipx, int clipy, int clipw, int cliph, PixelBuffer buf) {
        for(int i=0; i<text.length(); i++) {
            // FIXME: clipping
            char c = text.charAt(i);
            Glyph g = Glyph.getGlyph(font, fontsize, c);
	    buf.drawPicture(g.p,
                            x + hpad,
                            y + vpad + g.max_ascent - g.baseline,
                            x + hpad + g.p.getWidth(),
                            y + vpad + g.max_ascent - g.baseline + g.p.getHeight(),
                            0, 0,
                            g.p.getWidth(), g.p.getHeight());
            x += g.advance;
        }
    }


    // Methods to implement org.xwt.js.JS //////////////////////////////////////

    public Object callMethod(Object method, JS.Array args, boolean checkOnly) throws JS.Exn {
        if ("indexof".equals(method)) {
            if (checkOnly) return Boolean.TRUE;
            if (args.length() != 1 || args.elementAt(0) == null || !(args.elementAt(0) instanceof Box)) return new Integer(-1);
            Box b = (Box)args.elementAt(0);
            if (b.parent != Box.this) {
                if (redirect == null || redirect == Box.this) return new Integer(-1);
                return redirect.callMethod(method, args, checkOnly);
            }
            return new Integer(b.getIndexInParent());
        }
        return null;
    }

    /** Returns the i_th child */
    public Object get(int i) {
        if (redirect == null) return null;
        if (redirect != this) return redirect.get(i);
        return i >= numChildren() || i < 0 ? null : getChild(i);
    }

    /**
     *  Inserts value as child i; calls remove() if necessary.
     *  This method handles "reinserting" one of your children properly.
     *  INVARIANT: after completion, getChild(min(i, numChildren())) == newnode
     *  WARNING: O(n) runtime, unless i == numChildren()
     */
    public void put(int i, Object value) {
        if (i < 0) return;

        if (value != null && !(value instanceof Box)) {
            if (Log.on) Log.logJS(this, "attempt to set a numerical property on a box to anything other than a box");
        } else if (redirect == null) {
            if (Log.on) Log.logJS(this, "attempt to add/remove children to/from a node with a null redirect");
        } else if (redirect != this) {
            Box b = value == null ? (Box)redirect.get(i) : (Box)value;
            redirect.put(i, value);
            put("0", b);
        } else if (value == null) {
            if (i >= 0 && i < numChildren()) {
                Box b = getChild(i);
                b.remove();
                put("0", b);
            }
        } else {
            Box newnode = (Box)value;

            // check if box being moved is currently target of a redirect
            for(Box cur = newnode.parent; cur != null; cur = cur.parent)
                if (cur.redirect == newnode) {
                    if (Log.on) Log.logJS(this, "attempt to move a box that is the target of a redirect");
                    return;
                }

            // check for recursive ancestor violation
            for(Box cur = this; cur != null; cur = cur.parent)
                if (cur == newnode) {
                    if (Log.on) Log.logJS(this, "attempt to make a node a parent of its own ancestor");
                    if (Log.on) Log.log(this, "box == " + this + "  ancestor == " + newnode);
                    return;
                }

            if (numKids > 15 && children == null) convert_to_array();
            newnode.remove();
            newnode.parent = this;
            
            if (children == null) {
                if (firstKid == null) {
                    firstKid = newnode;
                    newnode.prevSibling = newnode;
                    newnode.nextSibling = newnode;
                } else if (i >= numKids) {
                    newnode.prevSibling = firstKid.prevSibling;
                    newnode.nextSibling = firstKid;
                    firstKid.prevSibling.nextSibling = newnode;
                    firstKid.prevSibling = newnode;
                } else {
                    Box cur = firstKid;
                    for(int j=0; j<i; j++) cur = cur.nextSibling;
                    newnode.prevSibling = cur.prevSibling;
                    newnode.nextSibling = cur;
                    cur.prevSibling.nextSibling = newnode;
                    cur.prevSibling = newnode;
                    if (i == 0) firstKid = newnode;
                }
                numKids++;
                
            } else {
                if (i >= children.size()) {
                    newnode.indexInParent = children.size();
                    children.addElement(newnode);
                } else {
                    children.insertElementAt(newnode, i);
                    for(int j=i; j<children.size(); j++)
                        getChild(j).indexInParent = j;
                }
            }
            
            // need both of these in case child was already uncalc'ed
            Box b = newnode; 
            MARK_FOR_REFLOW_b;
            MARK_FOR_REFLOW_this;
            
            newnode.dirty();

            // note that JavaScript box[0] will invoke put(int i), not put(String s)
            put("0", newnode);
        }
    }
    
    public Object get(Object key, Object key2) { return super.get(key, key2); }
    public void put(Object key, Object key2, Object val) { super.put(key, key2, val); }

    public Object get_(Object name) { return super.get(name); }
    public Object get(Object name) { return get(name, false); }
    public Object get(Object name_, boolean ignoretraps) {
        if (name_ instanceof Number) return get(((Number)name_).intValue());

        if (!(name_ instanceof String)) return null;
        String name = (String)name_;
        if (name.equals("")) return null;

        // See if we're triggering a trap
        Trap t = ignoretraps ? (Trap)null : (Trap)get(name, Trap.class);
        if (t != null) return t.perform();

        // Check for a special handler
        SpecialBoxProperty gph = (SpecialBoxProperty)SpecialBoxProperty.specialBoxProperties.get(name);
        if (gph != null) return gph.get(this);

        Object ret = super.get(name);
        if (name.startsWith("$") && ret == null)
            if (Log.on) Log.logJS(this, "WARNING: attempt to access " + name + ", but no child with id=\"" +
                                  name.substring(1) + "\" found");
        return ret;
    }

    public Object[] keys() {
        Object[] ret = new Object[numChildren()];
        for(int i=0; i<ret.length; i++) ret[i] = new Integer(i);
        return ret;
    }

    public void addTrap(Object name, Object value) { Trap.addTrap(this, name, ((JS.CompiledFunction)value)); }
    public void delTrap(Object name, Object value) { Trap.delTrap(this, name, ((JS.CompiledFunction)value)); }


    /**
     *  Scriptable.put()
     *  @param ignoretraps if set, no traps will be triggered (set when 'cascade' reaches the bottom of the trap stack)
     *  @param rp if this put is being performed via a root proxy, rp is the root proxy.
     */
    public void put_(Object name, Object value) { super.put(name, value); }
    public void put(Object name, Object value) { put(name, value, false); }
    public void put(Object name_, Object value, boolean ignoretraps) {
        if (name_ instanceof Number) { put(((Number)name_).intValue(), value); return; }
        if (!(name_ instanceof String)) { super.put(name_,value); return; }
        String name = name_.toString();
        if (!ignoretraps) {
            Trap t = (Trap)get(name, Trap.class);
            if (t != null) {
                t.perform(value);
                return;
            }
        }

        // don't want to really cascade down to the box on this one
        if (name.equals("0")) return;

        SpecialBoxProperty gph = (SpecialBoxProperty)SpecialBoxProperty.specialBoxProperties.get(name);
        if (gph != null) { gph.put(name, this, value); return; }

        super.put(name, value);
    }


    // Tree Manipulation /////////////////////////////////////////////////////////////////////

    /** The parent of this node */
    private Box parent = null;
    
    // Variables used in Vector mode */
    /** INVARIANT: if (parent != null) parent.children.elementAt(indexInParent) == this */
    private int indexInParent;
    private Vec children = null;

    // Variables used in linked-list mode
    private int numKids = 0;
    private Box nextSibling = null;
    private Box prevSibling = null;
    private Box firstKid = null;
    
    // when we get more than 15 children, we switch to array-mode
    private void convert_to_array() {
        children = new Vec(numKids);
        Box cur = firstKid;
        do {
            children.addElement(cur);
            cur.indexInParent = children.size() - 1;
            cur = cur.nextSibling;
        } while (cur != firstKid);
    }
    
    /** remove this node from its parent; INVARIANT: whenever the parent of a node is changed, remove() gets called. */
    public void remove() {
        if (parent == null) {
            if (surface != null) surface.dispose(true);
            return;
        }
        Box oldparent = parent;
        if (oldparent == null) return;
        MARK_FOR_REFLOW_this;
        dirty();
        flags &= ~MOUSEINSIDE_FLAG;

        if (parent.children != null) {
            parent.children.removeElementAt(indexInParent);
            for(int j=indexInParent; j<parent.children.size(); j++)
                (parent.getChild(j)).indexInParent = j;

        } else {
            if (parent.firstKid == this) {
                if (nextSibling == this) parent.firstKid = null;
                else parent.firstKid = nextSibling;
            }
            parent.numKids--;
            prevSibling.nextSibling = nextSibling;
            nextSibling.prevSibling = prevSibling;
            prevSibling = null;
            nextSibling = null;
        }
        parent = null;

        if (oldparent != null) { Box b = oldparent; MARK_FOR_REFLOW_b; }

        // note that JavaScript box[0] will invoke put(int i), not put(String s)
        if (oldparent != null) oldparent.put("0", this);
    }

    /** returns our next sibling (parent[ourindex + 1]) */
    public final Box nextSibling() {
        if (parent == null) return null;
        if (parent.children == null) {
            if (nextSibling == parent.firstKid) return null;
            return nextSibling;
        } else {
            if (indexInParent >= parent.children.size() - 1) return null;
            return (Box)parent.children.elementAt(indexInParent + 1);
        }
    }
    
    /** returns our next sibling (parent[ourindex + 1]) */
    public final Box prevSibling() {
        if (parent == null) return null;
        if (parent.children == null) {
            if (this == parent.firstKid) return null;
            return prevSibling;
        } else {
            if (indexInParent == 0) return null;
            return (Box)parent.children.elementAt(indexInParent - 1);
        }
    }
    
    /** Returns the parent of this node */
    public Box getParent() { return parent; }
    
    /** Returns ith child */
    public Box getChild(int i) {
        if (children == null) {
            if (firstKid == null) return null;
            if (i >= numKids) return null;
            if (i == numKids - 1) return firstKid.prevSibling;
            Box cur = firstKid;
            for(int j=0; j<i; j++) cur = cur.nextSibling;
            return cur;
        } else {
            if (i >= children.size() || i < 0) return null;
            return (Box)children.elementAt(i);
        }
    }
    
    /** Returns the number of children */
    public int numChildren() {
        if (children == null) {
            if (firstKid == null) return 0;
            int i=1;
            for(Box cur = firstKid.nextSibling; cur != firstKid; i++) cur = cur.nextSibling;
            return i;
        } else {
            return children.size();
        }
    }
    
    /** Returns our index in our parent */
    public int getIndexInParent() {
        if (parent == null) return 0;
        if (parent.children == null) {
            int i = 0;
            for(Box cur = this; cur != parent.firstKid; i++) cur = cur.prevSibling;
            return i;
        } else {
            return indexInParent;
        }
    }

    /** returns the root of the surface that this box belongs to */
    public final Box getRoot() {
        if (parent == null && surface != null) return this;
        if (parent == null) return null;
        return parent.getRoot();
    }


    // Trivial Helper Methods (should be inlined) /////////////////////////////////////////

    static final int min(int a, int b) { if (a<b) return a; else return b; }
    static final double min(double a, double b) { if (a<b) return a; else return b; }
    static final int max(int a, int b) { if (a>b) return a; else return b; }
    static final int min(int a, int b, int c) { if (a<=b && a<=c) return a; else if (b<=c && b<=a) return b; else return c; }
    static final int max(int a, int b, int c) { if (a>=b && a>=c) return a; else if (b>=c && b>=a) return b; else return c; }
    static final int bound(int a, int b, int c) { if (c < b) return c; if (a > b) return a; return b; }
    final boolean inside(int x, int y) { return (!((flags & INVISIBLE_FLAG) != 0) && x >= 0 && y >= 0 && x < width && y < height); }
    
    /** figures out what box in this subtree of the Box owns the pixel at x,y relitave to the Surface */
    public static Box whoIs(Box cur, int x, int y) {

        if (cur.parent != null) throw new Error("whoIs may only be invoked on the root box of a surface");
        int globalx = 0;
        int globaly = 0;

        // WARNING: this method is called from the event-queueing
        // thread -- it may run concurrently with ANY part of XWT, and
        // is UNSYNCHRONIZED for performance reasons.  BE CAREFUL
        // HERE.

        if ((cur.flags & INVISIBLE_FLAG) != 0) return null;
        if (!cur.inside(x - globalx, y - globaly)) return cur.parent == null ? cur : null;
        OUTER: while(true) {
            for(int i=cur.numChildren() - 1; i>=0; i--) {
                Box child = cur.getChild(i);
                if (child == null) continue;        // since this method is unsynchronized, we have to double-check
                globalx += child.x;
                globaly += child.y;
                if (!((child.flags & INVISIBLE_FLAG) != 0) && child.inside(x - globalx, y - globaly)) { cur = child; continue OUTER; }
                globalx -= child.x;
                globaly -= child.y;
            }
            break;
        }
        return cur;
    }

    /** 
     *  A helper class for properties of Box which require special
     *  handling.
     *
     *  To avoid excessive use of String.equals(), the Box.get() and
     *  Box.put() methods employ a Hash keyed on property names that
     *  require special handling. The value stored in the Hash is an
     *  instance of an anonymous subclass of SpecialBoxProperty, which knows
     *  how to handle get()s and put()s for that property name. There
     *  should be one anonymous subclass of SpecialBoxProperty for each
     *  specially-handled property on Box.
     */
    static class SpecialBoxProperty {

        SpecialBoxProperty() { }

        /** stores instances of SpecialBoxProperty; keyed on property name */
        static Hash specialBoxProperties = new Hash(200, 3);

        /** this method defines the behavior when the property is get()ed from b */
        Object get(Box b) { return null; }

        /** this method defines the behavior when the property is put() to b */
        void put(Box b, Object value) { }

        /** this method defines the behavior when the property is put() to b, allows a single SpecialBoxProperty to serve multiple properties */
        void put(String name, Box b, Object value) { put(b, value); }

        static {
            //#repeat fillcolor/strokecolor
            specialBoxProperties.put("fillcolor", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if ((b.fillcolor & 0xFF000000) == 0) return null;
                        String red = Integer.toHexString((b.fillcolor & 0x00FF0000) >> 16);
                        String green = Integer.toHexString((b.fillcolor & 0x0000FF00) >> 8);
                        String blue = Integer.toHexString(b.fillcolor & 0x000000FF);
                        if (red.length() < 2) red = "0" + red;
                        if (blue.length() < 2) blue = "0" + blue;
                        if (green.length() < 2) green = "0" + green;
                        return "#" + red + green + blue;
                    }
                    public void put(Box b, Object value) {
                        int newcolor = b.fillcolor;
                        String s = value == null ? null : value.toString();
                        if (value == null) newcolor = 0x00000000;
                        else if (s.length() > 0 && s.charAt(0) == '#')
                            try {
                                newcolor = 0xFF000000 |
                                    (Integer.parseInt(s.substring(1, 3), 16) << 16) |
                                    (Integer.parseInt(s.substring(3, 5), 16) << 8) |
                                    Integer.parseInt(s.substring(5, 7), 16);
                            } catch (NumberFormatException e) {
                                Log.log(this, "invalid color " + s);
                                return;
                            }
                        else if (org.xwt.translators.SVG.colors.get(s) != null)
                            newcolor = 0xFF000000 | ((Integer)org.xwt.translators.SVG.colors.get(s)).intValue();
                        if (newcolor == b.fillcolor) return;
                        b.fillcolor = newcolor;
                        b.dirty();
                    }
                });
            //#end
        
            specialBoxProperties.put("color", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.get("fillcolor"); }
                    public void put(Box b, Object value) { b.put("fillcolor", value); }
                });

            specialBoxProperties.put("textcolor", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.get("strokecolor"); }
                    public void put(Box b, Object value) { b.put("strokecolor", value); }
                });

            specialBoxProperties.put("text", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.text; }
                    public void put(Box b, Object value) {
                        String t = value == null ? "null" : value.toString();
                        if (t.equals(b.text)) return;
			b.text = t;
			if (t == null) {
                            if (b.textwidth != 0 || b.textheight != 0) MARK_FOR_REFLOW_b;
			    b.textwidth = b.textheight = 0;
			} else {
			    try {
                                MARK_FOR_REFLOW_b;
                                b.textwidth = 0;
                                for(int i=0; i<b.text.length(); i++) {
                                    Glyph g = Glyph.getGlyph(b.font, b.fontsize, b.text.charAt(i));
                                    b.textwidth += g.advance;
                                    b.textheight = g.max_ascent + g.max_descent;
                                }
			    } catch (Exception e) {
				Log.log(this, e);
			    }
			}
			b.dirty();
                    } });

            specialBoxProperties.put("font", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.font; }
                    public void put(Box b, Object value) {
                        // FIXME: translate value into a resource if it is a string
                        b.font = value == null ? null : (Res)value;
                        MARK_FOR_REFLOW_b;
                        b.flags |= FONT_CHANGED_FLAG;
                        b.dirty();
                    } });
        
            specialBoxProperties.put("fontsize", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.font; }
                    public void put(Box b, Object value) {
                        if (b.fontsize == stoi(value)) return;
                        b.fontsize = stoi(value);
                        MARK_FOR_REFLOW_b;
                        b.flags |= FONT_CHANGED_FLAG;
                        b.dirty();
                    } });
        
            specialBoxProperties.put("thisbox", new SpecialBoxProperty() {
                    public Object get(Box b) { return b; }
                    public void put(Box b, Object value) {
                        if (value == null) b.remove();
                        else if (value.equals("window") || value.equals("frame")) Platform.createSurface(b, value.equals("frame"), true);
                        else if (Log.on) Log.log(this, "put invalid value to 'thisbox' property: " + value);
                    }
                });

            specialBoxProperties.put("orient", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        Log.log(this, "warning: the orient property is deprecated");
                        if (b.redirect == null) return "horizontal";
                        else if (b.redirect != b) return get(b.redirect);
                        else if (b.cols == 1) return "vertical";
                        else if (b.rows == 1) return "horizontal";
                        else return "grid";
                    }
                    public void put(Box b, Object value) {
                        Log.log(this, "warning: the orient property is deprecated");
                        if (value == null) return;
                        if (b.redirect == null) return;
                        if (b.redirect != b) { put(b.redirect, value); return; }
                        if (value.equals("vertical")) {
                            if (b.rows == 0) return;
                            b.rows = 0; b.cols = 1;
                        } else if (value.equals("horizontal")) {
                            if (b.cols == 0) return;
                            b.cols = 0; b.rows = 1;
                        } else if (Log.on)
                            Log.log(this, "invalid value put to orient property: " + value);
                        MARK_FOR_REFLOW_b;
                    } });

            //FIXME
            specialBoxProperties.put("static", new SpecialBoxProperty() {
                });

            specialBoxProperties.put("shrink", new SpecialBoxProperty() {
                    public Object get(Box b) { return (((b.flags & HSHRINK_FLAG) != 0) || ((b.flags & VSHRINK_FLAG) != 0)) ? Boolean.TRUE : Boolean.FALSE; }
                    public void put(Box b, Object value) { b.put("hshrink", value); b.put("vshrink", value); }
                });
        
            //#repeat hshrink/vshrink HSHRINK_FLAG/VSHRINK_FLAG
            specialBoxProperties.put("hshrink", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Boolean((b.flags & HSHRINK_FLAG) != 0); }
                    public void put(Box b, Object value) {
                        boolean newshrink = stob(value);
                        if (((b.flags & HSHRINK_FLAG) != 0) == newshrink) return;
                        if (newshrink) b.flags |= HSHRINK_FLAG; else b.flags &= ~HSHRINK_FLAG;
                        MARK_FOR_REFLOW_b;
                    }
                });
            //#end
        
            //#repeat x/y
            specialBoxProperties.put("x", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if (b.surface == null) return new Integer(0);
                        if ((b.flags & INVISIBLE_FLAG) != 0) return new Integer(0);
                        return new Integer(b.x);
                    }
                    public void put(Box b, Object value) {
                        if (!((b.flags & ABSOLUTE_FLAG) != 0)) return;
                        int x = stoi(value);
                        if (x == b.x) return;
                        b.dirty();
                        b.x = x;
                        if (b.parent == null && b.surface != null) b.surface.setLocation();
                        MARK_FOR_REFLOW_b;
                        b.dirty();
                    }
                });
            //#end
        
            //#repeat width/height minwidth/minheight maxwidth/maxheight
            specialBoxProperties.put("width", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Integer(b.width); }
                    public void put(Box b, Object value) {
                        b.width = stoi(value);
                        if (b.parent == null && b.surface != null) {
                            b.surface.setSize();
                            MARK_FOR_REFLOW_b;
                        } else {
                            if (b.minwidth == b.width && b.maxwidth == b.width) return;
                            b.minwidth = b.maxwidth = b.width;
                            MARK_FOR_REFLOW_b;
                        }
                    } });
            //#end

            //#repeat cols/rows rows/cols
            specialBoxProperties.put("cols", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Double(b.cols); }
                    public void put(Box b, Object value) {
                        if (b.cols == stoi(value)) return;
                        b.cols = stoi(value);
                        if (b.cols == 0 && b.rows == 0) b.rows = 1;
                        if (b.cols != 0 && b.rows != 0) b.rows = 0;
                        MARK_FOR_REFLOW_b;
                    } });
            //#end
        
            //#repeat colspan/rowspan
            specialBoxProperties.put("colspan", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Double(b.colspan); }
                    public void put(Box b, Object value) {
                        if (b.colspan == stoi(value)) return;
                        b.colspan = stoi(value);
                        MARK_FOR_REFLOW_b;
                    }
                });
            //#end
        
            specialBoxProperties.put("tile", new SpecialBoxProperty() {
                    public Object get(Box b) { return ((b.flags & TILE_FLAG) != 0) ? Boolean.TRUE : Boolean.FALSE; }
                    public void put(Box b, Object value) {
                        if (((b.flags & TILE_FLAG) != 0) == stob(value)) return;
                        if (stob(value)) b.flags |= TILE_FLAG; else b.flags &= ~TILE_FLAG;
                        b.dirty();
                    } });
        
            specialBoxProperties.put("invisible", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        for (Box cur = b; cur != null; cur = cur.parent) {
                            if ((cur.flags & INVISIBLE_FLAG) != 0) return Boolean.TRUE; }
                        return Boolean.FALSE;
                    }
                    public void put(Box b, Object value) {
                        if (stob(value) == ((b.flags & INVISIBLE_FLAG) != 0)) return;
                        if (stob(value)) b.flags |= INVISIBLE_FLAG; else b.flags &= ~INVISIBLE_FLAG;
                        if (b.parent == null) {
                            if (b.surface != null) b.surface.setInvisible((b.flags & INVISIBLE_FLAG) != 0);
                        } else {
                            b.dirty();
                            MARK_FOR_REFLOW_b_parent;
                            b.parent.dirty(b.x, b.y, b.width, b.height);
                        }
                    }});
        
            specialBoxProperties.put("absolute", new SpecialBoxProperty() {
                    public Object get(Box b) { return ((b.flags & ABSOLUTE_FLAG) != 0) ? Boolean.TRUE : Boolean.FALSE; }
                    public void put(Box b, Object value) {
                        if (stob(value) == ((b.flags & ABSOLUTE_FLAG) != 0)) return;
                        if (stob(value)) b.flags |= ABSOLUTE_FLAG; else b.flags &= ~ABSOLUTE_FLAG;
                        if ((b.flags & ABSOLUTE_FLAG) != 0) { b.x = 0; b.y = 0; }
                        if (b.parent != null) MARK_FOR_REFLOW_b_parent;
                    } });
        
            specialBoxProperties.put("image", new SpecialBoxProperty() {
                    /* FIXME
                    public Object get(Box b) { return b.image == null ? null : ImageDecoder.imageToNameMap.get(b.image); }
                    */
                    public void put(Box b, Object value) {
                        //if (image != null) System.out.println("hit");
                        if (value == null) {
                            b.image = null;
                        } else if (value instanceof Res) {
                            b.image = Picture.fromRes((Res)value);
                        } else {
                            // FIXME
                        }
                        b.minwidth = min(b.maxwidth, max(b.minwidth, b.image == null ? 0 : b.image.getWidth()));
                        b.minheight = min(b.maxheight, max(b.minheight, b.image == null ? 0 : b.image.getHeight()));
                        MARK_FOR_REFLOW_b;
                        b.dirty();
                    }
                });

            //#repeat globalx/globaly x/y
            specialBoxProperties.put("globalx", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Integer(b.parent == null || b.surface == null ? 0 : b.x); }
                    public void put(Box b, Object value) {
                        if (b.surface == null || b.parent == null) return;
                        b.put("x", new Integer(stoi(value) - stoi(get(b.parent))));
                        MARK_FOR_REFLOW_b;
                    }
                });
            //#end
        
            specialBoxProperties.put("cursor", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.cursor; } 
                    public void put(Box b, Object value) {
                        b.cursor = (String)value;
                        if (b.surface == null) return;

                        // see if we need to update the surface cursor
                        Surface surface = b.getRoot().surface;
                        String tempcursor = surface.cursor;
                        b.Move(surface.mousex, surface.mousey, surface.mousex, surface.mousey);
                        if (surface.cursor != tempcursor) surface.syncCursor();
                    } 
                });
        
            //#repeat mousex/mousey x/y
            specialBoxProperties.put("mousex", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        Surface surface = b.getRoot().surface;
                        if (surface == null) return new Integer(0);
                        int mousex = surface.mousex;
                        for(Box cur = b; cur != null && cur.parent != null; cur = cur.parent) mousex -= cur.x;
                        return new Integer(mousex);
                    }
                });
            //#end
        
            specialBoxProperties.put("mouseinside", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        return ((b.flags & MOUSEINSIDE_FLAG) != 0) ? Boolean.TRUE : Boolean.FALSE; }
                });
        
            specialBoxProperties.put("numchildren", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if (b.redirect == null) return new Integer(0);
                        if (b.redirect != b) return get(b.redirect);
                        return new Integer(b.numChildren());
                    } });
        
            SpecialBoxProperty mouseEventHandler = new SpecialBoxProperty() {
                    public void put(String name, Box b, Object value) {
                        Surface surface = b.getRoot().surface;
                        if (surface == null) return;
                        int mousex = surface.mousex;
                        int mousey = surface.mousey;
                        for(Box c = b.parent; c != null && c.parent != null; c = c.parent) {
                            mousex -= c.x;
                            mousey -= c.y;
                        }
                        for(Box c = b.prevSibling(); c != null; c = c.prevSibling()) {
                            if (c.inside(mousex - c.x, mousey - c.y)) {
                                c.put(name, value);
                                return;
                            }
                        }
                        if (b.parent != null) b.parent.put(name, value);
                    }};

            specialBoxProperties.put("Press1", mouseEventHandler);
            specialBoxProperties.put("Press2", mouseEventHandler);
            specialBoxProperties.put("Press3", mouseEventHandler);
            specialBoxProperties.put("Release1", mouseEventHandler);
            specialBoxProperties.put("Release2", mouseEventHandler);
            specialBoxProperties.put("Release3", mouseEventHandler);
            specialBoxProperties.put("Click1", mouseEventHandler);
            specialBoxProperties.put("Click2", mouseEventHandler);
            specialBoxProperties.put("Click3", mouseEventHandler);
            specialBoxProperties.put("DoubleClick1", mouseEventHandler);
            specialBoxProperties.put("DoubleClick2", mouseEventHandler);
            specialBoxProperties.put("DoubleClick3", mouseEventHandler);

            specialBoxProperties.put("root", new SpecialBoxProperty() {
                    public void put(Box b, Object value) {
                        if (stob(value)) b.flags |= ISROOT_FLAG;
                        else b.flags &= ~ISROOT_FLAG;
                    }
                    public Object get(Box b) {
                        if (b.parent == null || ((b.flags & ISROOT_FLAG) != 0)) return b;
                        else return get(b.parent);
                    } });

            specialBoxProperties.put("Minimized", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if (b.parent == null && b.surface != null) return b.surface.minimized ? Boolean.TRUE : Boolean.FALSE;
                        else return null;
                    }
                    public void put(Box b, Object value) {
                        if (b.surface == null) return;
                        boolean val = stob(value);
                        if (b.parent == null && b.surface.minimized != val) b.surface.setMinimized(val);
                    }
                });

            specialBoxProperties.put("Maximized", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if (b.parent == null && b.surface != null) return b.surface.maximized ? Boolean.TRUE : Boolean.FALSE;
                        else return null;
                    }
                    public void put(Box b, Object value) {
                        if (b.surface == null) return;
                        boolean val = stob(value);
                        if (b.parent == null && b.surface.maximized != val) b.surface.setMaximized(val);
                    }
                });

            specialBoxProperties.put("toback", new SpecialBoxProperty() {
                    public void put(Box b, Object value) {
                        if (b.parent == null && stob(value) && b.surface != null) b.surface.toBack();
                    }
                });

            specialBoxProperties.put("tofront", new SpecialBoxProperty() {
                    public void put(Box b, Object value) {
                        if (b.parent == null && stob(value) && b.surface != null) b.surface.toFront();
                    }
                });

            specialBoxProperties.put("Close", new SpecialBoxProperty() {
                    public void put(Box b, Object value) {
                        if (b.parent == null && b.surface != null) b.surface.dispose(true);
                    }
                });

            // these are all do-nothings; just to prevent space from getting taken up in the params Hash.
            specialBoxProperties.put("KeyPressed", new SpecialBoxProperty());
            specialBoxProperties.put("KeyReleased", new SpecialBoxProperty());
            specialBoxProperties.put("PosChange", new SpecialBoxProperty());
            specialBoxProperties.put("SizeChange", new SpecialBoxProperty());

            //#repeat hpad/vpad 
            specialBoxProperties.put("hpad", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if (b.redirect == null) return new Integer(0);
                        if (b.redirect != b) return get(b.redirect);
                        return new Integer(b.hpad);
                    }
                    public void put(Box b, Object value) {
                        if (b.redirect == null) return;
                        if (b.redirect != b) { put(b.redirect, value); return; }
                        int newval = stoi(value);
                        if (newval == b.hpad) return;
                        b.hpad = newval;
                        MARK_FOR_REFLOW_b;
                    }
                });
            //#end

            //#repeat minwidth/minheight maxwidth/maxheight
            specialBoxProperties.put("minwidth", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Integer(b.minwidth); }
                    public void put(Box b, Object value) {
                        if (stoi(value) == b.minwidth) return;
                        b.minwidth = stoi(value);
                        MARK_FOR_REFLOW_b;
                    }
                });
            specialBoxProperties.put("maxwidth", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Integer(b.maxwidth); }
                    public void put(Box b, Object value) {
                        if (stoi(value) == b.maxwidth) return;
                        b.maxwidth = stoi(value);
                        MARK_FOR_REFLOW_b;
                    }
                });
            //#end

            specialBoxProperties.put("redirect", new SpecialBoxProperty() {
                    public void put(Box b, Object value) {
                        if (b.redirect != b) Log.log(this, "cannot change the redirect of a box once it is set");
                        else if (value == null) b.redirect = null;
                        else {
                            Box b2 = (Box)value;
                            while(b2 != b && b2 != null) b2 = b2.parent;
                            if (b2 == null) Log.log(this, "a box's redirect must be one of its descendants");
                            b.redirect = (Box)value;
                        }
                    }
                    public Object get(Box b) {
                        if (b.redirect == null) return null;
                        if (b.redirect == b) return Boolean.TRUE;
                        return get(b.redirect);
                    }
                });

            // FEATURE: this still isn't totally harmonious; when you createSurface, these aren't checked
            specialBoxProperties.put("titlebar", new SpecialBoxProperty() {
                    public void put(Box b, Object value) {
                        if (b.surface != null) b.surface.setTitleBarText(value.toString());
                        b.put_("titlebar", value);
                    }
                    public Object get(Box b) { return b.get_("titlebar"); }
                });

            specialBoxProperties.put("icon", new SpecialBoxProperty() {
                    // FIXME
                    /*
                    public void put(Box b, Object value) {
                        b.put_("icon", value);
                        if (b.surface == null) return;
                        Picture pic = ImageDecoder.getPicture(value.toString());
                        if (pic != null) b.surface.setIcon(pic);
                        else if (Log.on) Log.log(this, "unable to load icon " + value);
                    }
                    public Object get(Box b) { return b.get_("icon"); }
                    */
                });
        }
    }

    /** helper that converts a String to a boolean according to JavaScript coercion rules */
    public static boolean stob(Object o) {
        if (o == null) return false;
        return Boolean.TRUE.equals(o) || "true".equals(o);
    }

    /** helper that converts a String to an int according to JavaScript coercion rules */
    public static int stoi(Object o) {
        if (o == null) return 0;
        if (o instanceof Integer) return ((Integer)o).intValue();
        
        String s;
        if (!(o instanceof String)) s = o.toString();
        else s = (String)o;
        
        try { return Integer.parseInt(s.indexOf('.') == -1 ? s : s.substring(0, s.indexOf('.'))); }
        catch (NumberFormatException e) { return 0; }
    }
}
        

    
