// Copyright 2003 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import java.lang.reflect.*;
import java.net.*;
import java.io.*;
import java.util.*;
import org.xwt.util.*;

/** 
 *  Abstracts away the small irregularities in JVM implementations.
 *
 *  The default Platform class supports a vanilla JDK 1.1
 *  JVM. Subclasses are provided for other VMs. Methods whose names
 *  start with an underscore are meant to be overridden by
 *  subclasses. If you create a subclass of Platform, you should put
 *  it in the org.xwt.plat package, and add code to this file's static
 *  block to detect the new platform.
 */
public class Platform {

    // Static Data /////////////////////////////////////////////////////////////////////////////////////

    /**
     *  set to true during the delivery of a KeyPressed:C-v/A-v or Press3; it is safe to use a
     *  'global' here, since message delivery is single-threaded and non-preemptable
     */
    static boolean clipboardReadEnabled = false;

    /** The appropriate Platform object for this JVM */
    static Platform platform = null;

    /** true if proxy autodetection has already been run */
    static boolean alreadyDetectedProxy = false;

    /** the result of proxy autodetection */
    static HTTP.Proxy cachedProxyInfo = null;

    /** the current build */
    public static String build = "unknown";

    // VM Detection Logic /////////////////////////////////////////////////////////////////////

    /** do-nothing method that forces &lt;clinit&gt; to run */
    public static void forceLoad() { }

    // If you create a new subclass of Platform, you should add logic
    // here to detect it. Do not reference your class directly -- use
    // reflection.

    static {
        System.err.println("Detecting JVM...");
        try {
            String vendor = System.getProperty("java.vendor", "");
            String version = System.getProperty("java.version", "");
            String os_name = System.getProperty("os.name", "");
            String os_version = System.getProperty("os.version", "");
            String platform_class = null;
            
            if (os_name.startsWith("Darwin")) platform_class = "Darwin";
            else if (vendor.startsWith("Free Software Foundation")) {
                if (os_name.startsWith("Window")) platform_class = "Win32";
                else platform_class = "X11";
            } else if (version.startsWith("1.1") && vendor.startsWith("Netscape")) platform_class = "Netscape";
            else if (version.startsWith("1.1") && vendor.startsWith("Microsoft")) platform_class = "Microsoft";
            else if (!version.startsWith("1.0") && !version.startsWith("1.1")) platform_class = "Java2";

            /*
        // Disable 2d hardware acceleration on Jaguar
        if (os_name.equals("Mac OS X") && os_version.equals("10.2"))
            */
        System.setProperty("com.apple.hwaccel", "false");

            if (platform_class != null) {
                platform = (Platform)Class.forName("org.xwt.plat." + platform_class).newInstance();
                platform.init();
            }

            try {
                build = (String)Class.forName("org.xwt.Build").getField("build").get(null);
            } catch (ClassNotFoundException cnfe) {
            } catch (Exception e) {
                if (Log.on) Log.log(Platform.class, "exception while detecting build:");
                if (Log.on) Log.log(Platform.class, e);
            }
            if (Log.on) Log.log(Platform.class, "XWT build: " + build);

            if (Log.on) Log.log(Platform.class, "XWT VM detection:   vendor = " + vendor);
            if (Log.on) Log.log(Platform.class, "                   version = " + version);
            if (Log.on) Log.log(Platform.class, "                        os = " + os_name + " [version " + os_version + "]");

            if (platform_class == null) {
                if (Log.on) Log.log(Platform.class, "Unable to detect JVM");
                new Platform().criticalAbort("Unable to detect JVM");
            }

            if (Log.on) Log.log(Platform.class, "                  platform = " + platform.getDescriptiveName());
            if (Log.on) Log.log(Platform.class, "                     class = " + platform.getClass().getName());
            platform.postInit();

        } catch (Exception e) {
            if (Log.on) Log.log(Platform.class, "Exception while trying to detect JVM");
            if (Log.on) Log.log(Platform.class, e);
            new Platform().criticalAbort("Unable to detect JVM");
        }

    }


    // Methods to be Overridden ////////////////////////////////////////////////////////////////////

    /** a string describing the VM */
    protected String getDescriptiveName() { return "Generic Java 1.1 VM"; }

    /** this initializes the platform; code in here can invoke methods on Platform since Platform.platform has already been set */
    protected void init() { }
    protected void postInit() { }

    /** creates and returns a doublebuffer 'belonging' to <tt>owner</tt>; we need to associate PixelBuffers to surfaces
     *  due to AWT 1.1 requirements (definately for Navigator, possibly also for MSJVM).
     */
    public static PixelBuffer createPixelBuffer(int w, int h, Surface s) { return platform._createPixelBuffer(w, h, s); }
    protected PixelBuffer _createPixelBuffer(int w, int h, Surface owner) { return null; }
    
    /** creates and returns a picture */
    public static Picture createPicture(int[] data, int w, int h) { return platform._createPicture(data, w, h); }
    protected Picture _createPicture(int[] b, int w, int h) { return null; }

    /** creates a socket object */
    public static Socket getSocket(String host, int port, boolean ssl, boolean negotiate) throws IOException {
        return platform._getSocket(host, port, ssl, negotiate);
    }
    protected Socket _getSocket(String host, int port, boolean ssl, boolean negotiate) throws IOException {
        Socket ret = ssl ? new TinySSL(host, port, negotiate) : new Socket(java.net.InetAddress.getByName(host), port);
        ret.setTcpNoDelay(true);
        return ret;
    }
    
    /** should return true if it is safe to supress full-surface dirties immediately after a window resize */
    public static boolean supressDirtyOnResize() { return platform._supressDirtyOnResize(); }
    protected boolean _supressDirtyOnResize() { return false; }

    /** the human-readable name of the key mapped to XWT's 'alt' key */
    public static String altKeyName() { return platform._altKeyName(); }
    protected String _altKeyName() { return "alt"; }

    /** returns the contents of the clipboard */    
    public static Object getClipBoard() { return clipboardReadEnabled ? platform._getClipBoard() : null; }
    protected String _getClipBoard() { return null; }

    /** sets the contents of the clipboard */
    public static void setClipBoard(String s) { platform._setClipBoard(s); }
    protected void _setClipBoard(String s) { }

    /** returns the width of the screen, in pixels */
    public static int getScreenWidth() { return platform._getScreenWidth(); }
    protected int _getScreenWidth() { return 640; }

    /** returns the height of the screen, in pixels */
    public static int getScreenHeight() { return platform._getScreenHeight(); }
    protected int _getScreenHeight() { return 480; }

    /** returns the maximum number of threads that the XWT engine can create without adversely affecting the host OS */
    public static int maxThreads() { return platform._maxThreads(); }
    protected int _maxThreads() { return 25; }

    /** used to notify the user of very serious failures; usually used when logging is not working or unavailable */
    protected void _criticalAbort(String message) { System.exit(-1); }

    /** if true, org.xwt.Surface will generate a Click automatically after a press and a release */
    public static boolean needsAutoClick() { return platform._needsAutoClick(); }
    protected boolean _needsAutoClick() { return false; }

    /** if true, org.xwt.Surface will generate a DoubleClick automatically after recieving two clicks in a short period of time */
    public static boolean needsAutoDoubleClick() { return platform._needsAutoDoubleClick(); }
    protected boolean _needsAutoDoubleClick() { return false; }

    /** returns true iff the platform has a case-sensitive filesystem */
    public static boolean isCaseSensitive() { return platform._isCaseSensitive(); }
    protected boolean _isCaseSensitive() { return true; }

    /** returns an InputStream to the builtin xwar */
    public static InputStream getBuiltinInputStream() { return platform._getBuiltinInputStream(); }
    protected InputStream _getBuiltinInputStream() {
        return this.getClass().getClassLoader().getResourceAsStream("org/xwt/builtin.jar");
    }

    /** returns the value of the environment variable key, or null if no such key exists */
    public static String getEnv(String key) { return platform._getEnv(key); }
    protected String _getEnv(String key) {
        try {
            String os = System.getProperty("os.name").toLowerCase();
            Process p;
            if (os.indexOf("windows 9") != -1 || os.indexOf("windows me") != -1) {
                // hack -- jdk1.2/1.3 on Win32 pop open an ugly DOS box; 1.4 does not
                if (platform.getClass().getName().endsWith("Java12")) return null;
                p = Runtime.getRuntime().exec("command.com /c set");
            } else if (os.indexOf("windows") > -1) {
                // hack -- jdk1.2/1.3 on Win32 pop open an ugly DOS box; 1.4 does not
                if (platform.getClass().getName().endsWith("Java12")) return null;
                p = Runtime.getRuntime().exec("cmd.exe /c set");
            } else {  
                p = Runtime.getRuntime().exec("env");
            }
            BufferedReader br = new BufferedReader(new InputStreamReader(p.getInputStream()));
            String s;
            while ((s = br.readLine()) != null)
                if (s.startsWith(key + "="))
                    return s.substring(key.length() + 1);
        } catch (Exception e) {
            if (Log.on) Log.log(this, "Exception while reading from environment:");
            if (Log.on) Log.log(this, e);
        }
        return null;
    }

    /** convert a JPEG into an Image */
    public static synchronized Picture decodeJPEG(InputStream is, String name) { return platform._decodeJPEG(is, name); }
    protected Picture _decodeJPEG(InputStream is, String name) { return null; }

    /** displays a platform-specific "open file" dialog and returns the chosen filename, or null if the user hit cancel */
    protected String _fileDialog(String suggestedFileName, boolean write) { return null; }
    public static String fileDialog(String suggestedFileName, boolean write) {
        if (!ThreadMessage.suspendThread()) return null;
        try {
            return platform._fileDialog(suggestedFileName, write);
        } finally {
            ThreadMessage.resumeThread();
        }
    }

    /** default implementation is Eric Albert's BrowserLauncher.java */
    protected void _newBrowserWindow(String url) {
        try {
            Class c = Class.forName("edu.stanford.ejalbert.BrowserLauncher");
            Method m = c.getMethod("openURL", new Class[] { String.class });
            m.invoke(null, new String[] { url });
        } catch (Exception e) {
            Log.log(this, e);
        }
    }

    /** opens a new browser window */
    public static void newBrowserWindow(String url) {
        if (!(url.startsWith("https://") || url.startsWith("http://") || url.startsWith("ftp://") || url.startsWith("mailto:"))) {
            if (Log.on) Log.log(Platform.class, "xwt.newBrowserWindow() only supports http and https urls");
            return;
        }

        // check the URL for well-formedness, as a defense against buffer overflow attacks
        try {
            String u = url;
            if (u.startsWith("https")) u = "http" + u.substring(5);
            new URL(u);
        } catch (MalformedURLException e) {
            if (Log.on) Log.log(Platform.class, "URL " + url + " is not well-formed");
            if (Log.on) Log.log(Platform.class, e);
        }

        if (Log.on) Log.log(Platform.class, "newBrowserWindow, url = " + url);
        platform._newBrowserWindow(url);
    }

    /** used to notify the user of very serious failures; usually used when logging is not working or unavailable */
    public static void criticalAbort(String message) {
        if (Log.on) Log.log(Platform.class, "Critical Abort:");
        if (Log.on) Log.log(Platform.class, message);
        platform._criticalAbort(message);
    }

    /** this method invokes the platform _createSurface() method and then enforces a few post-call invariants */
    protected Surface _createSurface(Box b, boolean framed) { return null; }
    public static Surface createSurface(Box b, boolean framed, boolean refreshable) {
        Surface ret = platform._createSurface(b, framed);
        ret.setInvisible(false);

        Object titlebar = b.get("titlebar", true);
        if (titlebar != null) ret.setTitleBarText(titlebar.toString());

        Object icon = b.get("icon", true);
        if (icon != null && icon instanceof Res) {
            Picture pic = Picture.fromRes((Res)icon);
            if (pic != null) ret.setIcon(pic);
            else if (Log.on) Log.log(Platform.class, "unable to load icon " + icon);
        }

        ret.setLimits(b.minwidth, b.minheight, b.maxwidth, b.maxheight);

        if (refreshable) {
            Surface.allSurfaces.addElement(ret);
            ret.dirty(0, 0, b.width, b.height);
            ret.Refresh();
        }
        return ret;
    }

    /** detects proxy settings */
    protected synchronized HTTP.Proxy _detectProxy() { return null; }
    public static synchronized HTTP.Proxy detectProxy() {

        if (cachedProxyInfo != null) return cachedProxyInfo;
        if (alreadyDetectedProxy) return null;
        alreadyDetectedProxy = true;

        if (Log.on) Log.log(Platform.class, "attempting environment-variable DNS proxy detection");
        cachedProxyInfo = HTTP.Proxy.detectProxyViaManual();
        if (cachedProxyInfo != null) return cachedProxyInfo;

        if (Log.on) Log.log(Platform.class, "attempting " + platform.getClass().getName() + " proxy detection");
        cachedProxyInfo = platform._detectProxy();
        if (cachedProxyInfo != null) return cachedProxyInfo;

        return cachedProxyInfo;
    }

}


