// Copyright 2003 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import org.bouncycastle.util.encoders.Base64;
import org.xwt.util.*;
import java.io.*;
import java.util.*;

/** 
 *  A Surface, as described in the XWT Reference.
 *
 *  Platform subclasses should include an inner class subclass of
 *  Surface to return from the Platform._createSurface() method
 *
 *  Note that the members in the section 'state variables' are either
 *  in real-time (the actual size/position/state), or in
 *  MessageQueue-time (the size/position/state at the time that the
 *  now-executing message was enqueued). This distinction is important.
 */
// FIXME: put the scar box back in
public abstract class Surface extends PixelBuffer {

    public int getWidth() { return root == null ? 0 : root.width; }
    public int getHeight() { return root == null ? 0 : root.height; }
        
    // Static Data ////////////////////////////////////////////////////////////////////////////////

    /**< the most recently enqueued Move message; used to throttle the message rate */
    private static Message lastMoveMessage = null;

    /** all instances of Surface which need to be refreshed by the MessageQueue */
    public static Vec allSurfaces = new Vec();
    
    /** When set to true, render() should abort as soon as possible and restart the rendering process */
    static volatile boolean abort = false;

    public static boolean alt = false;          ///< true iff the alt button is pressed down, in real time
    public static boolean control = false;      ///< true iff the control button is pressed down, in real time
    public static boolean shift = false;        ///< true iff the shift button is pressed down, in real time
    public static boolean button1 = false;      ///< true iff button 1 is depressed, in MessageQueue-time
    public static boolean button2 = false;      ///< true iff button 2 is depressed, in MessageQueue-time
    public static boolean button3 = false;      ///< true iff button 3 is depressed, in MessageQueue-time

     

    // Instance Data ///////////////////////////////////////////////////////////////////////

    public Box root;      /**< The Box at the root of this surface */
    public String cursor = "default";

    public int mousex;                    ///< the x position of the mouse, relative to this Surface, in MessageQueue-time
    public int mousey;                    ///< the y position of the mouse, relative to this Surface, in MessageQueue-time
    public boolean minimized = false;     ///< True iff this surface is minimized, in real time
    public boolean maximized = false;     ///< True iff this surface is maximized, in real time

    /** Dirty regions on the backbuffer which need to be rebuilt using Box.render() */
    private DirtyList dirtyRegions = new DirtyList();


    // Used For Simulating Clicks and DoubleClicks /////////////////////////////////////////////////

    int last_press_x = Integer.MAX_VALUE;      ///< the x-position of the mouse the last time a Press message was enqueued
    int last_press_y = Integer.MAX_VALUE;      ///< the y-position of the mouse the last time a Press message was enqueued
    static int lastClickButton = 0;            ///< the last button to recieve a Click message; used for simulating DoubleClick's
    static long lastClickTime = 0;             ///< the last time a Click message was processed; used for simulating DoubleClick's
    
    
    // Methods to be overridden by subclasses ///////////////////////////////////////////////////////

    public abstract void toBack();      ///< when invoked, the surface should push itself to the back of the stacking order
    public abstract void toFront();     ///< when invoked, the surface should pull itself to the front of the stacking order
    public abstract void syncCursor();  ///< the <i>actual</i> cursor for this surface to the cursor referenced by <tt>cursor</tt>
    public abstract void setInvisible(boolean b);      ///< If <tt>b</tt>, make window invisible; otherwise, make it non-invisible.
    protected abstract void _setMaximized(boolean b);  ///< If <tt>b</tt>, maximize the surface; otherwise, un-maximize it.
    protected abstract void _setMinimized(boolean b);  ///< If <tt>b</tt>, minimize the surface; otherwise, un-minimize it.
    protected abstract void setSize(int width, int height);  ///< Sets the surface's width and height.
    public abstract void setLocation();                      ///< Set the surface's x/y position to that of the root box
    public abstract void setTitleBarText(String s);      ///< Sets the surface's title bar text, if applicable
    public abstract void setIcon(Picture i);      ///< Sets the surface's title bar text, if applicable
    public abstract void _dispose();      ///< Destroy the surface
    public void setLimits(int min_width, int min_height, int max_width, int max_height) { }


    // Helper methods for subclasses ////////////////////////////////////////////////////////////

    protected final void Press(final int button) {
        last_press_x = mousex;
        last_press_y = mousey;

        if (button == 1) button1 = true;
        else if (button == 2) button2 = true;
        else if (button == 3) button3 = true;

        if (button == 1) new SimpleMessage("Press1", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 2) new SimpleMessage("Press2", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 3) {
            final Box who = Box.whoIs(root, mousex, mousey);
            Message.Q.add(new Message() { public void perform() {
                Platform.clipboardReadEnabled = true;
                root.put("Press3", Boolean.TRUE);
                Platform.clipboardReadEnabled = false;
            }});
        }
    }

    protected final void Release(int button) {
        if (button == 1) button1 = false;
        else if (button == 2) button2 = false;
        else if (button == 3) button3 = false;

        if (button == 1) new SimpleMessage("Release1", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 2) new SimpleMessage("Release2", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 3) new SimpleMessage("Release3", Boolean.TRUE, Box.whoIs(root, mousex, mousey));

        if (Platform.needsAutoClick() && Math.abs(last_press_x - mousex) < 5 && Math.abs(last_press_y - mousey) < 5) Click(button);
        last_press_x = Integer.MAX_VALUE;
        last_press_y = Integer.MAX_VALUE;
    }

    protected final void Click(int button) {
        if (button == 1) new SimpleMessage("Click1", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 2) new SimpleMessage("Click2", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 3) new SimpleMessage("Click3", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        if (Platform.needsAutoDoubleClick()) {
            long now = System.currentTimeMillis();
            if (lastClickButton == button && now - lastClickTime < 350) DoubleClick(button);
            lastClickButton = button;
            lastClickTime = now;
        }
    }

    protected final void DoubleClick(int button) {
        if (button == 1) new SimpleMessage("DoubleClick1", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 2) new SimpleMessage("DoubleClick2", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 3) new SimpleMessage("DoubleClick3", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
    }

    /** sends a KeyPressed message; subclasses should not add the C- or A- prefixes, nor should they capitalize alphabet characters */
    protected final void KeyPressed(String key) {
        if (key == null) return;

        if (key.toLowerCase().endsWith("shift")) shift = true;
        else if (shift) key = key.toUpperCase();

        if (key.toLowerCase().equals("alt")) alt = true;
        else if (alt) key = "A-" + key;

        if (key.toLowerCase().endsWith("control")) control = true;
        else if (control) key = "C-" + key;

        final String fkey = key;
        Message.Q.add(new KMessage(key));
    }

    // This is implemented as a private static class instead of an anonymous class to work around a GCJ bug
    private class KMessage implements Message {
        String key = null;
        public KMessage(String k) { key = k; }
        public void perform() {
            if (key.equals("C-v") || key.equals("A-v")) Platform.clipboardReadEnabled = true;
            /* FIXME
            outer: for(int i=0; i<keywatchers.size(); i++) {
                Box b = (Box)keywatchers.elementAt(i);
                for(Box cur = b; cur != null; cur = cur.getParent())
                    if ((cur.flags & cur.INVISIBLE_FLAG) != 0) continue outer;
                b.put("KeyPressed", key);
            }
            */
            Platform.clipboardReadEnabled = false;
        }
    }

    /** sends a KeyReleased message; subclasses should not add the C- or A- prefixes, nor should they capitalize alphabet characters */
    protected final void KeyReleased(final String key) {
        if (key == null) return;
        if (key.toLowerCase().equals("alt")) alt = false;
        else if (key.toLowerCase().equals("control")) control = false;
        else if (key.toLowerCase().equals("shift")) shift = false;
        Message.Q.add(new Message() { public void perform() {
            /* FIXME
            outer: for(int i=0; i<keywatchers.size(); i++) {
                Box b = (Box)keywatchers.elementAt(i);
                for(Box cur = b; cur != null; cur = cur.getParent())
                    if ((cur.flags & cur.INVISIBLE_FLAG) != 0) continue outer;
                b.put("KeyReleased", key);
            }
            */
        }});
    }

    /**
     *  Notify XWT that the mouse has moved. If the mouse leaves the
     *  surface, but the host windowing system does not provide its new
     *  position (for example, a Java MouseListener.mouseExited()
     *  message), the subclass should use (-1,-1).
     */
    protected final void Move(final int newmousex, final int newmousey) {
        Message.Q.add(lastMoveMessage = new Message() { public void perform() {
            synchronized(Surface.this) {

                // if move messages are arriving faster than we can process them, we just start ignoring them
                if (lastMoveMessage != this) return;

                int oldmousex = mousex;
                int oldmousey = mousey;
                mousex = newmousex;
                mousey = newmousey;

                String oldcursor = cursor;
                cursor = "default";

                // Root gets motion events outside itself (if trapped, of course)
                if (!root.inside(oldmousex, oldmousey) && !root.inside(mousex, mousey) && (button1 || button2 || button3))
                    root.put("Move", Boolean.TRUE);

                root.Move(oldmousex, oldmousey, mousex, mousey);
                if (!cursor.equals(oldcursor)) syncCursor();
            }
        }});
    }

    protected final void SizeChange(final int width, final int height) {
        Message.Q.add(new Message() { public void perform() {
            if (width == root.width && height == root.height) return;
            root.needs_reflow = true;
            root.reflow(width, height);
        }});
        abort = true;
    }

    protected final void PosChange(final int x, final int y) {
        Message.Q.add(new Message() { public void perform() {
            root.put("x", new Integer(x));
            root.put("y", new Integer(y));
        }});
    }

    protected final void Close() { new SimpleMessage("Close", Boolean.TRUE, root); }
    protected final void Minimized(boolean b) { minimized = b; new SimpleMessage("Minimized", b ? Boolean.TRUE : Boolean.FALSE, root); }
    protected final void Maximized(boolean b) { maximized = b; new SimpleMessage("Maximized", b ? Boolean.TRUE : Boolean.FALSE, root); }
    protected final void Focused(boolean b) { new SimpleMessage("Focused", b ? Boolean.TRUE : Boolean.FALSE, root); }
    public static void Refresh() { Message.Q.refresh(); }

    public final void setMaximized(boolean b) { if (b != maximized) _setMaximized(maximized = b); }
    public final void setMinimized(boolean b) { if (b != minimized) _setMinimized(minimized = b); }


    // Other Methods ///////////////////////////////////////////////////////////////////////////////

    /** wrapper for setSize() which makes sure to dirty the place where the scar used to be */
    void setSize() {
        root.width = Math.max(root.width, 10);
        root.height = Math.max(root.height, 10);
        setSize(root.width, root.height);
    }

    /** Indicates that the Surface is no longer needed */
    public final void dispose(boolean quitIfAllSurfacesGone) {
        if (Log.on) Log.log(this, "disposing " + this);
        allSurfaces.removeElement(this);
        _dispose();
        if (allSurfaces.size() == 0) {
            if (Log.on) Log.log(this, "exiting because last surface was destroyed");
            System.exit(0);
        }
    }

    public void dirty(int x, int y, int w, int h) {
        dirtyRegions.dirty(x, y, w, h);
        Refresh();
    }

    public Surface(Box root) {

        this.root = root;
        if (root.surface != null && root.surface.root == root) root.surface.dispose(false);
        else root.remove();
        root.surface = this;

        // make sure the root is properly sized
        do { abort = false; root.reflow(); } while(abort);

        root.dirty();
        Refresh();
    }

    /** runs the prerender() and render() pipelines in the root Box to regenerate the backbuffer, then blits it to the screen */
    public synchronized void render() {

        // make sure the root is properly sized
        do {
            abort = false;
            root.reflow();
            // update mouseinside and trigger Enter/Leave as a result of box size/position changes
            String oldcursor = cursor;
            cursor = "default";
            root.Move(mousex, mousey, mousex, mousey);
            if (!cursor.equals(oldcursor)) syncCursor();
        } while(abort);

        Box.sizePosChangesSinceLastRender = 0;
        int[][] dirt = dirtyRegions.flush();
        for(int i = 0; dirt != null && i < dirt.length; i++) {
            if (dirt[i] == null) continue;
            int x = dirt[i][0], y = dirt[i][1], w = dirt[i][2], h = dirt[i][3];
            if (x < 0) x = 0;
            if (y < 0) y = 0;
            if (x+w > root.width) w = root.width - x;
            if (y+h > root.height) h = root.height - y;
            if (w <= 0 || h <= 0) continue;

            root.render(0, 0, x, y, w, h, this);
            
            if (abort) {

                // x,y,w,h is only partially reconstructed, so we must be careful not to re-blit it
                dirtyRegions.dirty(x, y, w, h);

                // put back all the dirty regions we haven't yet processed (including the current one)
                for(int j=i; j<dirt.length; j++)
                    if (dirt[j] != null)
                        dirtyRegions.dirty(dirt[j][0], dirt[j][1], dirt[j][2], dirt[j][3]);

                // tail-recurse
                render();
                return;
            }
        }

    }

    // FEATURE: reinstate recycler
    public class SimpleMessage implements Message {
        
        private Box boxContainingMouse;
        private Object value;
        public String name;
        
        SimpleMessage(String name, Object value, Box boxContainingMouse) {
            this.boxContainingMouse = boxContainingMouse;
            this.name = name;
            this.value = value;
            Message.Q.add(this);
        }
        
        public void perform() { boxContainingMouse.put(name, value); }
        public String toString() { return "SimpleMessage [name=" + name + ", value=" + value + "]"; }

    }


    // Default PixelBuffer implementation /////////////////////////////////////////////////////////

    public static abstract class DoubleBufferedSurface extends Surface {

        public DoubleBufferedSurface(Box root) { super(root); }
        PixelBuffer backbuffer = Platform.createPixelBuffer(Platform.getScreenWidth(), Platform.getScreenHeight(), this);
        DirtyList screenDirtyRegions = new DirtyList();

        public void drawPicture(Picture source, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2) {
            screenDirtyRegions.dirty(dx1, dy1, dx2 - dx1, dy2 - dy1);
            backbuffer.drawPicture(source, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2); }

        public void fillRect(int x1, int y1, int x2, int y2, int color) {
            screenDirtyRegions.dirty(x1, y1, x2 - x1, y2 - y1);
            backbuffer.fillRect(x1, y1, x2, y2, color); }

        public void render() {
            super.render();
            render_();
        }
        public void render_() {
            int[][] dirt = screenDirtyRegions.flush();
            for(int i = 0; dirt != null && i < dirt.length; i++) {
                if (dirt[i] == null) continue;
                int x = dirt[i][0];
                int y = dirt[i][1];
                int w = dirt[i][2];
                int h = dirt[i][3];
                if (x < 0) x = 0;
                if (y < 0) y = 0;
                if (x+w > root.width) w = root.width - x;
                if (y+h > root.height) h = root.height - y;
                if (w <= 0 || h <= 0) continue;
                blit(backbuffer, x, y, x, y, w + x, h + y);
            }
        }

        /** This is how subclasses signal a 'shallow dirty', indicating that although the backbuffer is valid, the screen is not */
        public final void Dirty(int x, int y, int w, int h) {
            screenDirtyRegions.dirty(x, y, w, h);
            Refresh();
        }

        /** copies a region from the doublebuffer to this surface */
        public abstract void blit(PixelBuffer source, int sx, int sy, int dx, int dy, int dx2, int dy2);

    }

}
