// Copyright 2003 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import java.io.*;
import java.util.zip.*;
import java.util.*;
import java.lang.*;
import org.xwt.js.*;
import org.xwt.util.*;

/**
 *  Encapsulates a template node (the &lt;template/&gt; element of a
 *  .xwt file, or any child element thereof).
 *
 *  Note that the Template instance corresponding to the
 *  &lt;template/&gt; node carries all the header information -- hence
 *  some of the instance members are not meaningful on non-root
 *  Template instances. We refer to these non-root instances as
 *  <i>anonymous templates</i>.
 *
 *  See the XWT reference for information on the order in which
 *  templates are applied, attributes are put, and scripts are run.
 */
public class Template {

    // Instance Members ///////////////////////////////////////////////////////

    String id = null;                  ///< the id of this box
    String redirect = null;            ///< the id of the redirect target; only meaningful on a root node
    private String[] keys;             ///< keys to be "put" to instances of this template; elements correspond to those of vals
    private Object[] vals;             ///< values to be "put" to instances of this template; elements correspond to those of keys
    private Vec children = new Vec();  ///< during XML parsing, this holds the list of currently-parsed children; null otherwise
    private int numunits = -1;         ///< see numUnits(); -1 means that this value has not yet been computed

    private JS.CompiledFunction script = null;       ///< the script on this node
    private String fileName = "unknown";             ///< the filename this node came from; used only for debugging
    private Vec preapply = new Vec();                ///< templates that should be preapplied (in the order of application)


    // Instance Members that are only meaningful on root Template //////////////////////////////////////

    private JS.Scope staticScope = null;             ///< the scope in which the static block is executed
    private JS.CompiledFunction staticscript = null; ///< the script on the static node of this template, null already performed


    // Only used during parsing /////////////////////////////////////////////////////////////////

    private StringBuffer content = null; ///< during XML parsing, this holds partially-read character data; null otherwise
    private int content_start = 0;       ///< line number of the first line of <tt>content</tt>
    private int content_lines = 0;       ///< number of lines in <tt>content</tt>
    private int startLine = -1;          ///< the line number that this element starts on
    private final Res r;                 ///< the resource we came from


    // Static data/methods ///////////////////////////////////////////////////////////////////

    public static Template getTemplate(Res r) {
        try {
            r = r.addExtension(".xwt");
            if (r.t != null) return r.t;
            r.t = new Template(r);
            new TemplateHelper().parseit(r.getInputStream(), r.t);
            return r.t;
        } catch (Exception e) {
            if (Log.on) Log.log(r.t.fileName, e);
            return null;
        }
    }

    public static Res resolveStringToResource(String str, XWT xwt, boolean permitAbsolute) {
        // URL
        if (str.indexOf("://") != -1) {
            if (permitAbsolute) return Res.stringToRes(str);
            Log.log(Template.class, "absolute URL " + str + " not permitted here");
            return null;
        }

        // root-relative
        Res ret = xwt.rr;
        while(str.indexOf('.') != -1) {
            String path = str.substring(0, str.indexOf('.'));
            str = str.substring(str.indexOf('.') + 1);
            ret = (Res)ret.get(path);
        }
        ret = (Res)ret.get(str);
        return ret;
    }


    // Methods to apply templates ////////////////////////////////////////////////////////

    private Template(Res r) { this.r = r; }

    /** called before this template is applied or its static object can be externally referenced */
    JS.Scope getStatic() {
        if (staticScope == null) staticScope = new JS.Scope(null);
        if (staticscript == null) return staticScope;
        JS.CompiledFunction temp = staticscript;
        staticscript = null;
        temp.call(new JS.Array(), staticScope);
        return staticScope;
    }
    
    /** Applies the template to Box b
     *  @param pboxes a vector of all box parents on which to put $-references
     *  @param ptemplates a vector of the fileNames to recieve private references on the pboxes
     */
    void apply(Box b, JS.Callable c, XWT xwt) { apply(b, c, xwt, null); }
    void apply(Box b, JS.Callable callback, XWT xwt, PerInstantiationScope parentPis) {

        getStatic();

        if (id != null) parentPis.putDollar(id, b);
        for(int i=0; i<preapply.size(); i++) {
            Template t = getTemplate(resolveStringToResource((String)preapply.elementAt(i), xwt, false));
            if (t == null) throw new RuntimeException("unable to resolve resource " + preapply.elementAt(i));
            t.apply(b, callback, xwt);
        }

        PerInstantiationScope pis = new PerInstantiationScope(b, xwt, parentPis);
        for (int i=0; children != null && i<children.size(); i++) {
            Box kid = new Box();
            ((Template)children.elementAt(i)).apply(kid, callback, xwt, pis);
            b.put(b.numChildren(), kid);
        }

        if (script != null) script.call(new JS.Array(), pis);

        for(int i=0; keys != null && i<keys.length; i++)
            if (vals[i] instanceof String && ((String)vals[i]).charAt(0) == '$') b.put(keys[i], pis.get(vals[i]));
            else if (keys[i].equals("image")) b.put("image", resolveStringToResource((String)vals[i], xwt, true));
            else b.put(keys[i], vals[i]);
    }



    // XML Parsing /////////////////////////////////////////////////////////////////

    /** handles XML parsing; builds a Template tree as it goes */
    static final class TemplateHelper extends XML {

        TemplateHelper() { }

        private int state;
        private static final int STATE_INITIAL = 0;
        private static final int STATE_IN_XWT_NODE = 1;
        private static final int STATE_IN_TEMPLATE_NODE = 2;
        private static final int STATE_FINISHED_TEMPLATE_NODE = 3;

        private String nameOfHeaderNodeBeingProcessed;

        Vec nodeStack = new Vec();  ///< stack of Templates whose XML elements we have seen open-tags for but not close-tags
        Template t = null;          ///< the template we're currently working on

        /** parse an XML input stream, building a Template tree off of <tt>root</tt> */
        void parseit(InputStream is, Template root) throws XML.XMLException, IOException {
            state = STATE_INITIAL;
            nameOfHeaderNodeBeingProcessed = null;
            nodeStack.setSize(0);
            t = root;
            parse(new InputStreamReader(is)); 
        }

        public void startElement(XML.Element c) throws XML.SchemaException {
            switch(state) {
            case STATE_INITIAL:
                if (!"xwt".equals(c.localName)) throw new XML.SchemaException("root element was not <xwt>");
                if (c.len != 0) throw new XML.SchemaException("root element must not have attributes");
                state = STATE_IN_XWT_NODE;
                return;

            case STATE_IN_XWT_NODE:
                if (nameOfHeaderNodeBeingProcessed != null) throw new XML.SchemaException("can't nest header nodes");
                nameOfHeaderNodeBeingProcessed = c.localName;
                if (c.localName.equals("doc")) {
                    // FIXME: ignore
                } else if (c.localName.equals("static")) {
                    if (t.staticscript != null)
                        throw new XML.SchemaException("the <static> header node may not appear more than once");
                    if (c.len > 0)
                        throw new XML.SchemaException("the <static> node may not have attributes");
                } else if (c.localName.equals("template")) {
                    t.startLine = getLine();
                    state = STATE_IN_TEMPLATE_NODE;
                    processBodyElement(c);
                } else {
                    throw new XML.SchemaException("unrecognized header node \"" + c.localName + "\"");
                }
                return;

            case STATE_IN_TEMPLATE_NODE:
                // push the last node we were in onto the stack
                nodeStack.addElement(t);
                // instantiate a new node, and set its fileName/importlist/preapply
                Template t2 = new Template(t.r);
                t2.startLine = getLine();
                if (!c.localName.equals("box") && !c.localName.equals("template"))
                    t2.preapply.addElement((c.uri == null ? "" : (c.uri + ".")) + c.localName);
                // make the new node the current node
                t = t2;
                processBodyElement(c);
                return;

            case STATE_FINISHED_TEMPLATE_NODE:
                throw new XML.SchemaException("no elements may appear after the <template> node");
            }
        }        

        private void processBodyElement(XML.Element c) {
            Hash h = new Hash(c.len * 2, 3);
            for(int i=0; i<c.len; i++) {
                if (c.keys[i].endsWith(":image")) {
                    String uri = (String)c.urimap.get(c.keys[i].substring(0, c.keys[i].indexOf(':')));
                    c.keys[i] = c.keys[i].substring(c.keys[i].lastIndexOf(':') + 1);
                    c.vals[i] = uri + "." + c.vals[i];
                }                    
                if ((c.keys[i].equals("preapply") || c.keys[i].endsWith(":preapply")) && c.localName.equals("template")) {
                    String uri = "";
                    if (c.keys[i].endsWith(":preapply")) {
                        uri = "." + c.urimap.get(c.keys[i].substring(0, c.keys[i].indexOf(':')));
                        c.keys[i] = c.keys[i].substring(c.keys[i].lastIndexOf(':') + 1);
                    }
                    StringTokenizer tok = new StringTokenizer(c.vals[i].toString(), " ");
                    while(tok.hasMoreTokens()) t.preapply.addElement(uri + tok.nextToken());
                    c.keys[i] = c.keys[c.keys.length - 1];
                    c.vals[i] = c.vals[c.vals.length - 1];
                    i--;
                    continue;
                }
                h.put(c.keys[i], c.vals[i]);
            }
            t.keys = new String[h.size()];
            t.vals = new Object[h.size()];

            Vec v = new Vec(h.size(), c.keys);
            v.sort(new Vec.CompareFunc() { public int compare(Object a, Object b) { return ((String)a).compareTo((String)b); } });
            for(int i=0; i<h.size(); i++) {
                if (c.keys[i].equals("thisbox")) {
                    for(int j=i; j>0; j--) { t.keys[j] = t.keys[j - 1]; t.vals[j] = t.vals[j - 1]; }
                    t.keys[0] = (String)v.elementAt(i);
                    t.vals[0] = h.get(t.keys[0]);
                } else {
                    t.keys[i] = (String)v.elementAt(i);
                    t.vals[i] = h.get(t.keys[i]);
                }
            }

            for(int i=0; i<t.keys.length; i++) {
                if (t.keys[i].equals("id")) {
                    t.id = t.vals[i].toString().intern();
                    t.keys[i] = null;
                    continue;
                }

                t.keys[i] = t.keys[i].intern();

                String valString = t.vals[i].toString();
                
                if (valString.equals("true")) t.vals[i] = Boolean.TRUE;
                else if (valString.equals("false")) t.vals[i] = Boolean.FALSE;
                else if (valString.equals("null")) t.vals[i] = null;
                else {
                    boolean hasNonNumeral = false;
                    boolean periodUsed = false;
                    for(int j=0; j<valString.length(); j++)
                        if (j == 0 && valString.charAt(j) == '-') {
                        } else if (valString.charAt(j) == '.' && !periodUsed && j != valString.length() - 1) {
                            periodUsed = true;
                        } else if (!Character.isDigit(valString.charAt(j))) {
                            hasNonNumeral = true;
                            break;
                        }
                    if (valString.length() > 0 && !hasNonNumeral) t.vals[i] = new Double(valString);
                    else t.vals[i] = valString.intern();
                }
            }
        }

        private JS.CompiledFunction parseScript(boolean isstatic) {
            JS.CompiledFunction thisscript = null;
            try {
                thisscript = JS.parse(t.fileName + (isstatic ? "._" : ""), t.content_start, new StringReader(t.content.toString()));
            } catch (IOException ioe) {
                if (Log.on) Log.log(this, "  ERROR: " + ioe.getMessage());
                thisscript = null;
            }
            t.content = null;
            t.content_start = 0;
            t.content_lines = 0;
            return thisscript;
        }

        public void endElement(XML.Element c) throws XML.SchemaException {
            if (state == STATE_IN_XWT_NODE) {
                if ("static".equals(nameOfHeaderNodeBeingProcessed) && t.content != null) t.staticscript = parseScript(true);
                nameOfHeaderNodeBeingProcessed = null;
                
            } else if (state == STATE_IN_TEMPLATE_NODE) {
                if (t.content != null) t.script = parseScript(false);
                if (nodeStack.size() == 0) {
                    // </template>
                    state = STATE_FINISHED_TEMPLATE_NODE;
                    
                } else {
                    // add this template as a child of its parent
                    Template oldt = t;
                    t = (Template)nodeStack.lastElement();
                    nodeStack.setSize(nodeStack.size() - 1);
                    t.children.addElement(oldt);
                }
            }
         }

        public void characters(char[] ch, int start, int length) throws XML.SchemaException {
            // invoke the no-tab crusade
            for (int i=0; length >i; i++) if (ch[start+i] == '\t') throw new XML.SchemaException(
                t.fileName+ ":" + getLine() + "," + getCol() + ": tabs are not allowed in XWT files");

            if ("static".equals(nameOfHeaderNodeBeingProcessed) || state == STATE_IN_TEMPLATE_NODE) {
                if (t.content == null) {
                    t.content_start = getLine();
                    t.content_lines = 0;
                    t.content = new StringBuffer();
                }

                t.content.append(ch, start, length);
                t.content_lines++;

            } else if (nameOfHeaderNodeBeingProcessed != null && state != STATE_FINISHED_TEMPLATE_NODE) {
                throw new XML.SchemaException("header node <" + nameOfHeaderNodeBeingProcessed + "> cannot have text content");
            }
        }

        public void whitespace(char[] ch, int start, int length) throws XML.SchemaException { }
    }

    private static class PerInstantiationScope extends JS.Scope {
        XWT xwt = null;
        PerInstantiationScope parentBoxPis = null;
        void putDollar(String key, Box target) {
            if (parentBoxPis != null) parentBoxPis.putDollar(key, target);
            declare("$" + key);
            put("$" + key, target);
        }
        public PerInstantiationScope(Scope parentScope, XWT xwt, PerInstantiationScope parentBoxPis) {
            super(parentScope);
            this.parentBoxPis = parentBoxPis;
            this.xwt = xwt;
        }
        public boolean isTransparent() { return true; }
        public boolean has(Object key) { return false; }
        public void declare(String s) { super.declare(s); }
        public Object get(Object key) {
            if (key.equals("xwt")) return xwt;
            if (Box.SpecialBoxProperty.specialBoxProperties.get(key.toString()) != null) return getParentScope().get(key);
            if (super.has(key)) return super.get(key);
            Object ret = xwt.rr.get(key);
            if (ret != null) return ret;
            throw new JS.Exn("must declare " + key + " before using it!");
        }
        public void put(Object key, Object val) {
            if (Box.SpecialBoxProperty.specialBoxProperties.get(key.toString()) != null) getParentScope().put(key, val);
            else if (super.has(key)) super.put(key, val);
            else throw new JS.Exn("must declare " + key + " before using it!");
        }
    }

}


