// Copyright 2003 Adam Megacz, see the COPYING file for licensing [GPL] 

package org.xwt.js; 
import org.xwt.util.*; 
import java.io.*;
import java.util.*;

/**
 *  The public API for the JS engine.  JS itself is actually a class
 *  implementing the absolute minimal amount of functionality for an
 *  Object which can be manipulated by JavaScript code.  The static
 *  methods, fields, and inner classes of JS define the publicly
 *  visible API for the XWT JavaScript engine; code outside this
 *  package should never depend on anything not defined in this file.
 */
public abstract class JS { 


    // Public Helper Methods //////////////////////////////////////////////////////////////////////

    /** parse and compile a function */
    public static CompiledFunction parse(String sourceName, int firstLine, Reader sourceCode) throws IOException {
        return new CompiledFunction(sourceName, firstLine, sourceCode, null);
    }

    /** coerce an object to a Boolean */
    public static boolean toBoolean(Object o) {
        if (o == null) return false;
        if (o instanceof Boolean) return ((Boolean)o).booleanValue();
        if (o instanceof Long) return ((Long)o).longValue() != 0;
        if (o instanceof Integer) return ((Integer)o).intValue() != 0;
        if (o instanceof Number) {
            double d = ((Number) o).doubleValue();
            // NOTE: d == d is a test for NaN. It should be faster than Double.isNaN()
            return d != 0.0 && d == d;
        }
        if (o instanceof String) return ((String)o).length() != 0;
        return true;
    }

    /** coerce an object to a Long */
    public static long toLong(Object o) { return toNumber(o).longValue(); }

    /** coerce an object to an Int */
    public static int toInt(Object o) { return toNumber(o).intValue(); }

    /** coerce an object to a Double */
    public static double toDouble(Object o) { return toNumber(o).doubleValue(); }

    /** coerce an object to a Number */
    public static Number toNumber(Object o) {
        if (o == null) return new Long(0);
        if (o instanceof Number) return ((Number)o);

        // NOTE: There are about 3 pages of rules in ecma262 about string to number conversions
        //       We aren't even close to following all those rules.  We probably never will be.
        if (o instanceof String) try { return new Double((String)o); } catch (NumberFormatException e) { return new Double(Double.NaN); }
        if (o instanceof Boolean) return ((Boolean)o).booleanValue() ? new Long(1) : new Long(0);
        if (o instanceof JS) return ((JS)o).coerceToNumber();
        throw new Error("toNumber() got object of type " + o.getClass().getName() + " which we don't know how to handle");
    }
    
    /** coerce an object to a String */
    public static String toString(Object o) {
        if(o == null) return "null";
        if(o instanceof String) return (String) o;
        if(o instanceof Integer || o instanceof Long || o instanceof Boolean) return o.toString();
        if(o instanceof JS) return ((JS)o).coerceToString();
        if(o instanceof Double || o instanceof Float) {
            double d = ((Number)o).doubleValue();
            if((int)d == d) return Integer.toString((int)d);
            return o.toString();
        }
        return o.toString();
    }
    
    // Instance Methods ////////////////////////////////////////////////////////////////////
 
    public abstract Object get(Object key) throws JS.Exn; 
    public abstract void put(Object key, Object val) throws JS.Exn; 
    public abstract Object[] keys(); 
    public Object callMethod(Object method, Array args, boolean checkOnly) throws JS.Exn {
	if(checkOnly) return Boolean.FALSE;
	Object o = get(method);
	if(o instanceof JS.Callable) {
	    return ((JS.Callable)o).call(args);
	} else if(o == null) {
	    throw new JS.Exn("Attempted to call non-existent method: " + method);
	} else {
	    throw new JS.Exn("Attempted to call a non-method: " +method);
	}
    }
    
    public Number coerceToNumber() { return new Integer(0); }
    public String coerceToString() { throw new JS.Exn("tried to coerce a JavaScript object to a String"); }
    public boolean coerceToBoolean() { return true; }
    
    public String typeName() { return "object"; }


    // Inner Classes /////////////////////////////////////////////////////////////////////////

    /** A sensible implementation of the abstract methods in the JS class */
    public static class Obj extends JS {
        private Hash entries = null;
        private boolean sealed = false;
        public Obj() { this(false); }
        public Obj(boolean sealed) { this.sealed = sealed; }
        public void setSeal(boolean sealed) { this.sealed = sealed; }      ///< a sealed object cannot have its properties modified
        public void put(Object key, Object val) { put(key, null, val); }
        protected void put(Object key, Object key2, Object val) {
            if (sealed) return;
            if (entries == null) entries = new Hash();
            entries.put(key, key2, val); }
        public Object[] keys() { return entries == null ? new Object[0] : entries.keys(); }
        public Object get(Object key) { return get(key, null); }
        protected Object get(Object key, Object key2) {
            if (entries == null) return null;
            if(key2 == null && callMethod((String)key, null, true) == Boolean.TRUE)
                return new Internal.CallableStub(this, key);
            return entries.get(key, key2);
        }
    }

    /** An exception which can be thrown and caught by JavaScript code */
    public static class Exn extends RuntimeException { 
        private Object js = null; 
        public Exn(Object js) { this.js = js; } 
        public String toString() { return "JS.Exn: " + js; }
        public String getMessage() { return toString(); }
        public Object getObject() { return js; } 
    } 

    /** The publicly-visible face of JavaScript Array objects */
    public static class Array extends ArrayImpl {
        public Array() { }
        public Array(int size) { super(size); }
        public void setSize(int i) { super.setSize(i); }
        public int length() { return super.length(); }
        public Object elementAt(int i) { return super.elementAt(i); }
        public void addElement(Object o) { super.addElement(o); }
        public void setElementAt(Object o, int i) { super.setElementAt(o, i); }
        public Object get(Object key) { return super._get(key); }
        public void put(Object key, Object val) { super._put(key, val); }
    }

    /** Any object which becomes part of the scope chain must support this interface */ 
    public static class Scope extends ScopeImpl { 
        public Scope(Scope parentScope) { this(parentScope, false); }
        public Scope(Scope parentScope, boolean sealed) { super(parentScope, sealed); }
        /** transparent scopes are not returned by THIS */
        public boolean isTransparent() { return super.isTransparent(); }
        public boolean has(Object key) { return super.has(key); }
        public Object get(Object key) { return super._get(key); }
        public void put(Object key, Object val) { super._put(key, val); }
        public void declare(String s) { super.declare(s); }
    } 

    /** anything that is callable with the () operator */
    public static abstract class Callable extends JS.Obj {
        public abstract Object call(JS.Array args) throws JS.Exn;
    }

    /** a Callable which was compiled from JavaScript code */
    public static class CompiledFunction extends CompiledFunctionImpl {
        public int getNumFormalArgs() { return numFormalArgs; }
        CompiledFunction(String sourceName, int firstLine, Reader sourceCode, Scope scope) throws IOException {
            super(sourceName, firstLine, sourceCode, scope);
        }
    }
    
    /** a scope that is populated with js objects and functions normally found in the global scope */
    public static class GlobalScope extends GlobalScopeImpl {
        public GlobalScope() { this(null); }
        public GlobalScope(JS.Scope parent) { super(parent); }
    }

    public static final JS Math = new org.xwt.js.Math();
 
    /** encapsulates a single JavaScript thread; the JS.Thread->java.lang.Thread mapping is 1:1 */
    public static class Thread {

        CompiledFunction currentCompiledFunction = null;
        Vec stack = new Vec();
        int pc;

        /** binds this thread to the current Java Thread */
        public void bindToCurrentJavaThread() { javaThreadToJSThread.put(java.lang.Thread.currentThread(), this); }

        /** returns the line of code that is currently executing */
        public int getLine() { return currentCompiledFunction == null ? -1 : currentCompiledFunction.getLine(pc); }

        /** returns the name of the source code file which declared the currently executing function */
        public String getSourceName() { return currentCompiledFunction == null ? null : currentCompiledFunction.getSourceName();  }

        /** fetches the currently-executing javascript function */
        public JS.CompiledFunction getCurrentCompiledFunction() { return currentCompiledFunction; }


        // Statics ///////////////////////////////////////////////////////////////////////

        private static Hashtable javaThreadToJSThread = new Hashtable();

        /** returns the JS thread for a given Java thread, creating one if necessary */
        public static JS.Thread fromJavaThread(java.lang.Thread t) {
            JS.Thread ret = (JS.Thread)javaThreadToJSThread.get(t);
            if (ret == null) {
                ret = new JS.Thread();
                ret.bindToCurrentJavaThread();
            }
            return ret;
        }
        
        public static JS.Thread currentJSThread() {
            return fromJavaThread(java.lang.Thread.currentThread());
        }
    }
} 




