// Copyright 2003 Adam Megacz, see the COPYING file for licensing [LGPL]
package org.xwt.plat;

import gnu.gcj.RawData;
import java.net.*;
import java.lang.reflect.*;
import java.io.*;
import java.util.*;
import org.xwt.util.*;
import org.xwt.*;

/** Platform implementation for POSIX compliant operating systems with an X11 Server */
public class X11 extends POSIX {

    // Static Data ///////////////////////////////////////////////////////////

    /**
     *  When the user reads from the clipboard, the main thread blocks
     *  on this semaphore until we get an X11 SelectionNotify. Crude,
     *  but effective. We know that only one thread will ever block on
     *  this, since only one thread can ever be running JavaScript.
     */
    public static Semaphore waiting_for_selection_event = new Semaphore();

    /** our local (in-process) copy of the clipboard */
    public static String clipboard = null;

    /** map from Window's (casted to jlong, wrapped in java.lang.Long) to X11Surface objects */
    public static Hashtable windowToSurfaceMap = new Hashtable();


    // General Methods ///////////////////////////////////////////////////////

    protected String _getAltKeyName() { return System.getProperty("os.name", "").indexOf("SunOS") != -1 ? "Meta" : "Alt"; }

    protected Picture _createPicture(int[] data, int w, int h) { return new X11Picture(data, w, h); }
    protected PixelBuffer _createPixelBuffer(int w, int h, Surface owner) { return new X11PixelBuffer(w, h); }
    protected Surface _createSurface(Box b, boolean framed) { return new X11Surface(b, framed); }
    protected boolean _needsAutoClick() { return true; }
    protected native int _getScreenWidth();
    protected native int _getScreenHeight();
    protected native String _getClipBoard();
    protected native void _setClipBoard(String s);
    protected boolean _needsAutoDoubleClick() { return true; }
    protected native void eventThread();
    private native void natInit();

    public X11() { }
    public void init() {
        natInit();
        (new Thread() { public void run() { eventThread(); } }).start();
    }

    // X11Surface /////////////////////////////////////////////////////

    /** Implements a Surface as an X11 Window */
    public static class X11Surface extends Surface.DoubleBufferedSurface {
        
        gnu.gcj.RawData window;
        gnu.gcj.RawData gc;
        boolean framed = false;
        Semaphore waitForCreation = new Semaphore();
        
        public native void setInvisible(boolean i);
        public void _setMaximized(boolean m) { if (Log.on) Log.log(this, "X11 can't maximize windows"); }
        public native void setIcon(Picture p);
        public native void _setMinimized(boolean b);
        public native void setTitleBarText(String s);
        public native void setSize(int w, int h);
        public native void setLocation();
        public native void natInit();
        public native void toFront();
        public native void toBack();
        public native void syncCursor();
        public native void _dispose();
        public native void setLimits(int minw, int minh, int maxw, int maxh);
        public native void blit(PixelBuffer s, int sx, int sy, int dx, int dy, int dx2, int dy2);
        public native void dispatchEvent(gnu.gcj.RawData ev);

        public X11Surface(Box root, boolean framed) {
            super(root);
            this.framed = framed;
            natInit();
        }        
    
    }


    // Our Subclass of Picture ///////////////////////////////////////////////

    /**
     *  Implements a Picture. No special X11 structure is created
     *  unless the image has no alpha (in which case a
     *  non-shared-pixmap PixelBuffer is created), or all-or-nothing
     *  alpha (in which case a non-shared-pixmap PixelBuffer with a
     *  stipple bitmap is created).
     */
    public static class X11Picture extends Picture {
        
        int width;
        int height;
        int[] data = null;
        public X11PixelBuffer doublebuf = null;

        public int getWidth() { return width; }
        public int getHeight() { return height; }

        public X11Picture(int[] data, int w, int h) {
            this.data = data;
            this.width = w;
            this.height = h;
            boolean needsStipple = false;

            // if we have any non-0x00, non-0xFF alphas, we can't double buffer ourselves
            for(int i=0; i<w*h; i++)
                if ((data[i] & 0xFF000000) == 0xFF000000)
                    needsStipple = true;
                else if ((data[i] & 0xFF000000) != 0x00)
                    return;

            buildPixelBuffer(needsStipple);
        }

        void buildPixelBuffer(boolean needsStipple) {
            if (doublebuf != null) return;
            // no point in using a shared pixmap since we'll only write to this image once
            X11PixelBuffer b = new X11PixelBuffer(width, height, false);
            b.drawPicture(this, 0, 0);
            if (needsStipple) b.createStipple(this);
            doublebuf = b;
        }

    }

    /**
     *  An X11PixelBuffer is implemented as an X11 pixmap. "Normal"
     *  PixelBuffers will use XShm shared pixmaps if
     *  available. X11PixelBuffers created to accelerate Pictures
     *  with all-or-nothing alpha will not use shared pixmaps, however
     *  (since they are only written to once.
     */
    public static class X11PixelBuffer extends PixelBuffer {

        int clipx, clipy, clipw, cliph;
        int width;
        int height;

        /** PixelBuffers of X11Pictures can have stipples -- the stipple of the Picture */
        RawData stipple = null;

        /** Sets the PixelBuffer's internal stipple to the alpha==0x00 regions of xpi */
        public native void createStipple(X11Picture xpi);
        
        RawData pm;                    // Pixmap (if any) representing this Picture
        boolean shared_pixmap = false; // true if pm is a ShmPixmap
        RawData fake_ximage = null;    // a 'fake' XImage corresponding to the shared pixmap; gives us the address and depth parameters
        RawData shm_segment = null;    // XShmSegmentInfo

        RawData gc;                    // Graphics Context on pm (never changes, so it's fast)
        RawData clipped_gc;            // Graphics Context on pm, use this one if you need a clip/stipple

        /** PixelBuffer mode */
        public X11PixelBuffer(int w, int h) { this(w, h, true); }
        public X11PixelBuffer(int w, int h, boolean shared_pixmap) {
            width = clipw = w;
            height = cliph = h;
            clipx = clipy = 0;
            this.shared_pixmap = shared_pixmap;
            natInit();
        }

        public void setClip(int x, int y, int x2, int y2) {
            clipx = x; if (clipx < 0) clipx = 0;
            clipy = y; if (clipy < 0) clipy = 0;
            clipw = x2 - x; if (clipw < 0) clipw = 0;
            cliph = y2 - y; if (cliph < 0) cliph = 0;
        }
        
        public void drawPicture(Picture source, int x, int y) {
            drawPicture(source, x, y, x + source.getWidth(), y + source.getHeight(), 0, 0, source.getWidth(), source.getHeight());
        }

        public void drawPicture(Picture source, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2) {
            if (!(dx2 - dx1 != sx2 - sx1 || dy2 - dy1 != sy2 - sy1) && ((X11Picture)source).doublebuf != null)
                fastDrawPicture(source, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2);
            else 
                slowDrawPicture(source, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2);
        }

        /** fast path for image drawing (no scaling, all-or-nothing alpha) */
        public native void fastDrawPicture(Picture source, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2);

        /** slow path for image drawing */
        public native void slowDrawPicture(Picture source, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2);

        public int getWidth() { return width; }
        public int getHeight() { return height; }
        public native void natInit();
        public native void fillRect(int x, int y, int x2, int y2, int color);
        public native void finalize();

    }

}
